# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=redefined-outer-name
from __future__ import annotations

import pytest

from capellambse.filehandler import memory

TEST_CONTENT = b"Hello, World!"


@pytest.fixture
def fh() -> memory.MemoryFileHandler:
    fh = memory.MemoryFileHandler()
    with fh.open("test.txt", "w") as f:
        f.write(TEST_CONTENT)
    return fh


@pytest.fixture
def fhdir() -> memory.MemoryFileHandler:
    fh = memory.MemoryFileHandler()
    fh.write_file("top.txt", TEST_CONTENT)
    fh.write_file("some/test.txt", TEST_CONTENT)
    fh.write_file("some/test2.txt", TEST_CONTENT)
    fh.write_file("other/test3.txt", TEST_CONTENT)
    return fh


def test_MemoryFileHandler_raises_ValueError_for_invalid_path():
    with pytest.raises(ValueError):
        memory.MemoryFileHandler(path="memory://invalid")


def test_MemoryFileHandler_raises_FileNotFoundError_for_nonexistent_file():
    fh = memory.MemoryFileHandler()
    with pytest.raises(FileNotFoundError):
        fh.open("test.txt")


def test_MemoryFileHandler_preserves_written_data(fh):
    with fh.open("test.txt", "r") as f:
        assert f.read() == TEST_CONTENT


def test_MemoryFiles_return_bytes_objects_from_read(fh):
    with fh.open("test.txt", "r") as f:
        assert isinstance(f.read(), bytes)


def test_empty_MemoryFileHandler_has_no_files():
    fh = memory.MemoryFileHandler()
    assert not [p.name for p in fh.iterdir()]


def test_MemoryFileHandler_iterates_over_files(fh):
    assert [p.name for p in fh.iterdir()] == ["test.txt"]


def test_MemoryFileHandler_iterated_files_can_be_read(fh):
    for p in fh.iterdir():
        with p.open("rb") as f:
            assert f.read() == TEST_CONTENT


def test_MemoryFilePath_returns_correct_name():
    fh = memory.MemoryFileHandler()
    assert fh.rootdir.joinpath("test.txt").name == "test.txt"


def test_MemoryFilePath_recognizes_files(fh):
    assert fh.rootdir.joinpath("test.txt").is_file()


def test_MemoryFilePath_recognizes_directories(fh):
    assert fh.rootdir.is_dir()


def test_MemoryFileHandler_can_iterate_subdirs(fhdir):
    actual = [p.name for p in fhdir.rootdir.iterdir("some")]

    assert actual == ["test.txt", "test2.txt"]


def test_MemoryFilePath_can_be_iterated(fhdir):
    actual = [p.name for p in fhdir.rootdir.joinpath("some").iterdir()]

    assert actual == ["test.txt", "test2.txt"]


def test_MemoryFilePath_can_iterate_over_subdirs(fhdir):
    actual = [p.name for p in fhdir.rootdir.iterdir("some")]

    assert actual == ["test.txt", "test2.txt"]



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=unused-variable
"""Static duck-typing assertions checked by mypy."""
from __future__ import annotations

from capellambse import model


def protocol_ModelObject_compliance():
    mobj: model.common.ModelObject

    mobj = model.GenericElement()  # type: ignore[call-arg]
    mobj = model.MelodyModel()  # type: ignore[call-arg]
    mobj = model.common.accessors._Specification()  # type: ignore[call-arg]



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import hashlib
import io
import pathlib
import shutil
import subprocess
import sys

import pytest

import capellambse
from capellambse import decl, helpers

# pylint: disable-next=relative-beyond-top-level, useless-suppression
from .conftest import (  # type: ignore[import-untyped]
    INSTALLED_PACKAGE,
    TEST_MODEL,
    TEST_ROOT,
)

DATAPATH = pathlib.Path(__file__).parent / "data" / "decl"
MODELPATH = pathlib.Path(TEST_ROOT / "5_0")

ROOT_COMPONENT = helpers.UUIDString("0d2edb8f-fa34-4e73-89ec-fb9a63001440")
ROOT_FUNCTION = helpers.UUIDString("f28ec0f8-f3b3-43a0-8af7-79f194b29a2d")


class TestDumpLoad:
    @staticmethod
    def test_promises_are_serialized_with_promise_tag():
        id = "some-future-object"
        data = [{"parent": decl.Promise(id)}]
        expected = f"- parent: !promise {id!r}\n"

        actual = decl.dump(data)

        assert actual == expected

    @staticmethod
    def test_uuid_references_are_serialized_with_uuid_tag():
        uuid = helpers.UUIDString("00000000-0000-0000-0000-000000000000")
        data = [{"parent": decl.UUIDReference(uuid)}]
        expected = f"- parent: !uuid {uuid!r}\n"

        actual = decl.dump(data)

        assert actual == expected

    @staticmethod
    def test_promise_tags_are_deserialized_as_promise():
        id = "some-future-object"
        yaml = f"- parent: !promise {id!r}\n"
        expected = [{"parent": decl.Promise(id)}]

        actual = decl.load(io.StringIO(yaml))

        assert actual == expected

    @staticmethod
    def test_uuid_tags_are_deserialized_as_uuidreference():
        uuid = helpers.UUIDString("00000000-0000-0000-0000-000000000000")
        yaml = f"- parent: !uuid {uuid!r}\n"
        expected = [{"parent": decl.UUIDReference(uuid)}]

        actual = decl.load(io.StringIO(yaml))

        assert actual == expected


class TestApplyExtend:
    @staticmethod
    def test_decl_errors_on_unknown_operations(
        model: capellambse.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              invalid_operation: {{}}
            """

        with pytest.raises(ValueError, match="invalid_operation"):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    @pytest.mark.parametrize(
        ["parent_str", "parent_getter"],
        [
            pytest.param(
                f"!uuid {ROOT_FUNCTION}",
                lambda m: m.by_uuid(ROOT_FUNCTION),
                id="!uuid",
            ),
        ],
    )
    def test_decl_finds_parent_to_act_on(
        model: capellambse.MelodyModel, parent_str, parent_getter
    ) -> None:
        funcname = "pass the unit test"
        yml = f"""\
            - parent: {parent_str}
              extend:
                functions:
                  - name: {funcname!r}
            """
        expected_len = len(model.search()) + 1
        assert funcname not in parent_getter(model).functions.by_name

        decl.apply(model, io.StringIO(yml))

        actual_len = len(model.search())
        assert actual_len == expected_len
        assert funcname in parent_getter(model).functions.by_name

    @staticmethod
    def test_decl_creates_each_object_in_a_list(
        model: capellambse.MelodyModel,
    ) -> None:
        parent_obj = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the first test
                  - name: pass the second test
                  - name: pass the third test
            """
        expected_len = len(model.search()) + 3
        for i in ("first", "second", "third"):
            assert f"pass the {i} test" not in parent_obj.functions.by_name

        decl.apply(model, io.StringIO(yml))

        actual_len = len(model.search())
        assert actual_len == expected_len
        for i in ("first", "second", "third"):
            assert f"pass the {i} test" in parent_obj.functions.by_name

    @staticmethod
    def test_decl_creates_nested_complex_objects_where_they_belong(
        model: capellambse.MelodyModel,
    ) -> None:
        root = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the unit test
                    functions:
                      - name: run the test function
                      - name: make assertions
            """
        expected_len = len(model.search()) + 3

        decl.apply(model, io.StringIO(yml))

        actual_len = len(model.search())
        assert actual_len == expected_len
        assert "pass the unit test" in root.functions.by_name
        parent = root.functions.by_name("pass the unit test", single=True)
        assert "run the test function" in parent.functions.by_name
        assert "make assertions" in parent.functions.by_name

    @staticmethod
    @pytest.mark.parametrize(
        "type",
        ["AttributeDefinition", "AttributeDefinitionEnumeration"],
    )
    def test_decl_can_disambiguate_creations_with_type_hints(
        model: capellambse.MelodyModel, type: str
    ) -> None:
        module_id = "db47fca9-ddb6-4397-8d4b-e397e53d277e"
        module = model.by_uuid(module_id)
        yml = f"""\
            - parent: !uuid {module_id}
              extend:
                attribute_definitions:
                  - long_name: New attribute
                    _type: {type}
            """
        assert "New attribute" not in module.attribute_definitions.by_long_name

        decl.apply(model, io.StringIO(yml))

        attrdefs = module.attribute_definitions
        assert "New attribute" in attrdefs.by_long_name
        newdef = attrdefs.by_long_name("New attribute", single=True)
        assert newdef.xtype.endswith(f":{type}")

    @staticmethod
    def test_decl_can_create_simple_objects_by_passing_plain_strings(
        model: capellambse.MelodyModel,
    ) -> None:
        attrdef_id = "637caf95-3229-4607-99a0-7d7b990bc97f"
        attrdef = model.by_uuid(attrdef_id)
        yml = f"""\
            - parent: !uuid {attrdef_id}
              extend:
                values:
                  - NewGame++
            """
        assert "NewGame++" not in attrdef.values

        decl.apply(model, io.StringIO(yml))

        assert "NewGame++" in attrdef.values

    @staticmethod
    def test_extend_operations_change_model_objects_in_place(
        model: capellambse.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        fnc = model.by_uuid("8833d2dc-b862-4a50-b26c-6f7e0f17faef")
        old_parent = fnc.parent
        assert old_parent != root_function
        assert fnc not in root_function.functions
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - !uuid {fnc.uuid}
            """

        decl.apply(model, io.StringIO(yml))

        assert fnc in root_function.functions
        assert root_function == fnc.parent

    @staticmethod
    def test_extend_operations_with_promises_change_model_objects_in_place(
        model: capellambse.MelodyModel,
    ) -> None:
        root_component = model.by_uuid(ROOT_COMPONENT)
        fnc_name = "The promised one"
        assert fnc_name not in root_component.functions.by_name
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: {fnc_name}
                    promise_id: promised-fnc
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                functions:
                  - !promise promised-fnc
            """

        decl.apply(model, io.StringIO(yml))

        assert fnc_name in root_component.functions.by_name

    @staticmethod
    def test_extend_operations_on_faulty_attribute_cause_an_exception(
        model: capellambse.MelodyModel,
    ) -> None:
        non_existing_attr = "Thought up"
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                {non_existing_attr}:
                  - !uuid {ROOT_FUNCTION}
            """

        with pytest.raises(TypeError, match=non_existing_attr):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    def test_promises_are_resolved_during_the_second_attempt(
        model: capellambse.MelodyModel,
    ) -> None:
        PHYS_COMPONENT = "b327d900-abd2-4138-a111-9ff0684739d8"
        cmp = model.by_uuid(PHYS_COMPONENT)
        previous_ports = len(cmp.ports)
        yml = f"""\
            - parent: !uuid {PHYS_COMPONENT}
              extend:
                physical_links:
                  - name: My new link
                    exchanges:
                      - !promise my_exchange
            - parent: !uuid {PHYS_COMPONENT}
              extend:
                exchanges:
                  - source: !promise first-port
                    target: !promise second-port
                    promise_id: my_exchange
            - parent: !uuid {PHYS_COMPONENT}
              extend:
                ports:
                  - name: First port
                    promise_id: first-port
                  - name: Second port
                    promise_id: second-port
            """

        decl.apply(model, io.StringIO(yml))

        assert len(cmp.ports) == previous_ports + 2
        ex = cmp.physical_links.by_name("My new link")
        assert len(ex.exchanges) == 1
        assert ex.exchanges[0].source.name == "First port"


class TestApplyPromises:
    @staticmethod
    def test_promises_can_backwards_reference_objects(
        model: capellambse.MelodyModel,
    ) -> None:
        root_func = model.by_uuid(ROOT_FUNCTION)
        root_comp = model.by_uuid(ROOT_COMPONENT)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the unit test
                    promise_id: pass-test
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise pass-test
            """
        expected_len = len(root_comp.allocated_functions) + 1

        decl.apply(model, io.StringIO(yml))

        actual_len = len(root_comp.allocated_functions)
        assert actual_len == expected_len
        assert "pass the unit test" in root_func.functions.by_name
        assert "pass the unit test" in root_comp.allocated_functions.by_name

    @staticmethod
    def test_promises_can_forward_reference_objects(
        model: capellambse.MelodyModel,
    ) -> None:
        root_func = model.by_uuid(ROOT_FUNCTION)
        root_comp = model.by_uuid(ROOT_COMPONENT)
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise pass-test
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the unit test
                    promise_id: pass-test
            """
        expected_len = len(root_comp.allocated_functions) + 1

        decl.apply(model, io.StringIO(yml))

        actual_len = len(root_comp.allocated_functions)
        assert actual_len == expected_len
        assert "pass the unit test" in root_func.functions.by_name
        assert "pass the unit test" in root_comp.allocated_functions.by_name

    @staticmethod
    @pytest.mark.parametrize(
        "order",
        [
            pytest.param((0, 1), id="backward"),
            pytest.param((1, 0), id="forward"),
        ],
    )
    def test_promises_on_simple_attributes_can_reference_objects(
        model: capellambse.MelodyModel, order
    ) -> None:
        root_func = model.by_uuid(ROOT_FUNCTION)
        snippets = (
            f"""
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                inputs:
                  - name: My input
                    promise_id: inport
                outputs:
                  - name: My output
                    promise_id: outport
            """,
            f"""
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                exchanges:
                  - name: Test exchange
                    source: !promise outport
                    target: !promise inport
            """,
        )
        yml = snippets[order[0]] + snippets[order[1]]

        decl.apply(model, io.StringIO(yml))

        assert "Test exchange" in root_func.exchanges.by_name
        exc = root_func.exchanges.by_name("Test exchange")
        assert exc.source == root_func.outputs.by_name("My output")
        assert exc.target == root_func.inputs.by_name("My input")

    @staticmethod
    def test_reused_promise_ids_cause_an_exception(
        model: capellambse.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: pass the first unit test
                    promise_id: colliding-promise-id
                  - name: pass the second unit test
                    promise_id: colliding-promise-id
            """

        with pytest.raises(ValueError, match=r"\bcolliding-promise-id\b"):
            decl.apply(model, io.StringIO(yml))

    @staticmethod
    def test_unfulfilled_promises_raise_an_exception(
        model: capellambse.MelodyModel,
    ) -> None:
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              extend:
                allocated_functions:
                  - !promise pass-test
            """

        with pytest.raises(decl.UnfulfilledPromisesError, match="^pass-test$"):
            decl.apply(model, io.StringIO(yml))


class TestApplyModify:
    @staticmethod
    def test_modify_operations_change_model_objects_in_place(
        model: capellambse.MelodyModel,
    ) -> None:
        newname = "Coffee machine"
        root_component = model.by_uuid(ROOT_COMPONENT)
        assert root_component.name != newname
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              modify:
                name: {newname}
            """

        decl.apply(model, io.StringIO(yml))

        assert root_component.name == newname

    @staticmethod
    def test_modify_can_set_attributes_to_promises(
        model: capellambse.MelodyModel,
    ) -> None:
        root_component = model.by_uuid(ROOT_COMPONENT)
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              modify:
                allocated_functions:
                  - !promise make-coffee
            - parent: !uuid {ROOT_FUNCTION}
              extend:
                functions:
                  - name: make coffee
                    promise_id: make-coffee
            """
        assert "make coffee" not in root_component.allocated_functions.by_name

        decl.apply(model, io.StringIO(yml))

        assert "make coffee" in root_component.allocated_functions.by_name

    @staticmethod
    def test_modify_can_set_attributes_to_uuid_references(
        model: capellambse.MelodyModel,
    ) -> None:
        root_component = model.by_uuid(ROOT_COMPONENT)
        root_function = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_COMPONENT}
              modify:
                allocated_functions:
                  - !uuid {ROOT_FUNCTION}
            """
        assert root_function not in root_component.allocated_functions

        decl.apply(model, io.StringIO(yml))

        assert root_function in root_component.allocated_functions

    @staticmethod
    def test_modifying_to_a_list_removes_all_previous_list_members(
        model: capellambse.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              modify:
                functions:
                  - name: survive
            """
        assert len(root_function.functions) > 0

        decl.apply(model, io.StringIO(yml))

        assert len(root_function.functions) == 1
        assert root_function.functions[0].name == "survive"


class TestApplyDelete:
    @staticmethod
    def test_delete_operations_remove_child_objects_from_the_model(
        model: capellambse.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        assert len(root_function.functions) > 0
        subfunc: str = root_function.functions[0].uuid
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              delete:
                functions:
                  - !uuid {subfunc}
            """
        assert subfunc in root_function.functions.by_uuid
        expected_len = len(root_function.functions) - 1

        decl.apply(model, io.StringIO(yml))

        assert len(root_function.functions) == expected_len
        assert subfunc not in root_function.functions.by_uuid
        with pytest.raises(KeyError, match=subfunc):
            model.by_uuid(subfunc)

    @staticmethod
    def test_delete_operations_delete_attributes_if_no_list_of_uuids_was_given(
        model: capellambse.MelodyModel,
    ) -> None:
        root_function = model.by_uuid(ROOT_FUNCTION)
        assert len(root_function.functions) > 0
        yml = f"""\
            - parent: !uuid {ROOT_FUNCTION}
              delete:
                functions:
            """

        decl.apply(model, io.StringIO(yml))

        assert len(root_function.functions) == 0


@pytest.mark.parametrize("filename", ["coffee-machine.yml"])
def test_full_example(model: capellambse.MelodyModel, filename: str):
    decl.apply(model, DATAPATH / filename)


def test_cli_applies_a_yaml_and_saves_the_model_back(tmp_path: pathlib.Path):
    shutil.copytree(MODELPATH, tmp_path / "model")
    model = tmp_path / "model" / TEST_MODEL
    semmodel = model.with_suffix(".capella")
    oldhash = hashlib.sha256(semmodel.read_bytes()).hexdigest()
    declfile = DATAPATH / "coffee-machine.yml"

    cli = subprocess.run(
        [sys.executable, "-mcapellambse.decl", f"--model={model}", declfile],
        cwd=INSTALLED_PACKAGE.parent,
        check=False,
    )

    assert cli.returncode == 0, "CLI process exited unsuccessfully"
    newhash = hashlib.sha256(semmodel.read_bytes()).hexdigest()
    assert newhash != oldhash, "Files on disk didn't change"



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The 'Requirements' namespace."""
from __future__ import annotations

__all__ = [
    "AbstractRequirementsAttribute",
    "AbstractRequirementsRelation",
    "AbstractType",
    "AttributeAccessor",
    "AttributeDefinition",
    "AttributeDefinitionEnumeration",
    "BooleanValueAttribute",
    "DataTypeDefinition",
    "DateValueAttribute",
    "EnumValue",
    "EnumerationDataTypeDefinition",
    "EnumerationValueAttribute",
    "Folder",
    "IntegerValueAttribute",
    "InternalRelation",
    "ModuleType",
    "RealValueAttribute",
    "RelationType",
    "ReqIFElement",
    "Requirement",
    "RequirementType",
    "StringValueAttribute",
]

import re

import markupsafe

import capellambse.model.common as c

c.XTYPE_ANCHORS[__name__] = "Requirements"


class ReqIFElement(c.GenericElement):
    """Attributes shared by all ReqIF elements."""

    identifier = c.AttributeProperty("ReqIFIdentifier", optional=True)
    long_name = c.AttributeProperty("ReqIFLongName", optional=True)
    description: str = c.AttributeProperty(  # type: ignore[assignment]
        "ReqIFDescription", optional=True
    )
    name = c.AttributeProperty("ReqIFName", optional=True)
    prefix = c.AttributeProperty("ReqIFPrefix", optional=True)
    type: c.Accessor = property(lambda _: None)  # type: ignore[assignment]

    def _short_repr_(self) -> str:  # pragma: no cover
        mytype = type(self).__name__
        parent = self._element
        if self.xtype in {  # FIXME remove this shit
            "CapellaRequirements:CapellaTypesFolder",
            "Requirements:DataTypeDefinition",
            "Requirements:RequirementType",
            "Requirements:RelationType",
            "Requirements:ModuleType",
            "Requirements:EnumerationDataTypeDefinition",
            "Requirements:EnumValue",
            "Requirements:AttributeDefinition",
            "Requirements:AttributeDefinitionEnumeration",
        }:
            return f'<{mytype} {parent.get("ReqIFLongName")!r} ({self.uuid})>'

        name = (
            parent.get("ReqIFName")
            or parent.get("ReqIFChapterName")
            or parent.get("ReqIFLongName")
            or ""
        )
        return f"<{mytype} {name!r} ({self.uuid})>"

    def _short_html_(self) -> markupsafe.Markup:
        name = markupsafe.Markup.escape(self.name or self.long_name)
        return markupsafe.Markup(self._wrap_short_html(f" &quot;{name}&quot;"))


@c.xtype_handler(None)
class DataTypeDefinition(ReqIFElement):
    """A data type definition for requirement types."""

    _xmltag = "ownedDefinitionTypes"


@c.xtype_handler(None)
class AttributeDefinition(ReqIFElement):
    """An attribute definition for requirement types."""

    _xmltag = "ownedAttributes"

    data_type = c.AttrProxyAccessor(DataTypeDefinition, "definitionType")


class AbstractRequirementsAttribute(c.GenericElement):
    _xmltag = "ownedAttributes"

    definition = c.AttrProxyAccessor(AttributeDefinition, "definition")

    value: c.AttributeProperty | c.AttrProxyAccessor

    def __repr__(self) -> str:
        return self._short_repr_()

    def _short_repr_(self) -> str:
        mytype = type(self).__name__
        if self.definition is not None:
            defname = self.definition.long_name
        else:
            defname = re.sub(r"(?<!^)([A-Z])", r" \1", mytype)
        return f"<{mytype} [{defname}] {self._repr_value()} ({self.uuid})>"

    def _short_html_(self) -> markupsafe.Markup:
        if self.definition is not None:
            name = markupsafe.Markup.escape(self.definition.long_name)
        else:
            name = markupsafe.Markup.escape("None")
        return markupsafe.Markup(self._wrap_short_html(f" &quot;{name}&quot;"))

    def _repr_value(self) -> str:
        return repr(self.value)


class AttributeAccessor(c.DirectProxyAccessor[AbstractRequirementsAttribute]):
    def __init__(self) -> None:
        super().__init__(
            c.GenericElement,  # type: ignore[arg-type]
            (
                BooleanValueAttribute,
                DateValueAttribute,
                EnumerationValueAttribute,
                IntegerValueAttribute,
                RealValueAttribute,
                StringValueAttribute,
            ),
            aslist=c.MixedElementList,
            list_extra_args={
                "mapkey": "definition.long_name",
                "mapvalue": "value",
            },
        )

    def _match_xtype(  # type: ignore[override]
        self, type_: str
    ) -> tuple[type, str]:
        type_ = type_.lower()
        try:
            cls = _attr_type_hints[type_]
            return cls, c.build_xtype(cls)
        except KeyError:
            raise ValueError(f"Invalid type hint given: {type_!r}") from None


@c.xtype_handler(None)
class BooleanValueAttribute(AbstractRequirementsAttribute):
    """A string value attribute."""

    value = c.BooleanAttributeProperty("value")


@c.xtype_handler(None)
class DateValueAttribute(AbstractRequirementsAttribute):
    """A value attribute that stores a date and time."""

    value = c.DatetimeAttributeProperty("value", optional=True)

    def _repr_value(self) -> str:
        if self.value is None:
            return "None"
        return self.value.isoformat()


@c.xtype_handler(None)
class IntegerValueAttribute(AbstractRequirementsAttribute):
    """An integer value attribute."""

    value = c.AttributeProperty("value", returntype=int, default=0)


@c.xtype_handler(None)
class RealValueAttribute(AbstractRequirementsAttribute):
    """A floating-point number value attribute."""

    value = c.AttributeProperty("value", returntype=float, default=0.0)


@c.xtype_handler(None)
class StringValueAttribute(AbstractRequirementsAttribute):
    """A string value attribute."""

    value = c.AttributeProperty("value", default="")


@c.xtype_handler(None)
@c.attr_equal("long_name")
class EnumValue(ReqIFElement):
    """An enumeration value for :class:`.EnumerationDataTypeDefinition`."""

    _xmltag = "specifiedValues"

    def __str__(self) -> str:
        return self.long_name


@c.xtype_handler(None)
class EnumerationDataTypeDefinition(ReqIFElement):
    """An enumeration data type definition for requirement types."""

    _xmltag = "ownedDefinitionTypes"

    values = c.DirectProxyAccessor(
        EnumValue, aslist=c.ElementList, single_attr="long_name"
    )


@c.xtype_handler(None)
class AttributeDefinitionEnumeration(ReqIFElement):
    """An enumeration attribute definition for requirement types."""

    _xmltag = "ownedAttributes"

    data_type = c.AttrProxyAccessor(
        EnumerationDataTypeDefinition, "definitionType"
    )
    multi_valued = c.BooleanAttributeProperty(
        "multiValued",
        __doc__=(
            "Boolean flag for setting multiple enumeration values on"
            " the attribute"
        ),
    )


@c.xtype_handler(None)
class EnumerationValueAttribute(AbstractRequirementsAttribute):
    """An enumeration attribute."""

    definition = c.AttrProxyAccessor(
        AttributeDefinitionEnumeration, "definition"  # type: ignore[arg-type]
    )
    values = c.AttrProxyAccessor(EnumValue, "values", aslist=c.ElementList)

    @property
    def value(self):
        vals = self.values
        if len(vals) > 1:
            raise TypeError("Multi-value enumeration, use `.values` instead")
        if len(vals) == 1:
            return vals[0]
        return None  # TODO return a default value?

    def _repr_value(self) -> str:
        return repr([i.long_name for i in self.values])


@c.attr_equal("long_name")
class AbstractType(ReqIFElement):
    owner = c.ParentAccessor(c.GenericElement)
    attribute_definitions = c.DirectProxyAccessor(
        c.GenericElement,
        (AttributeDefinition, AttributeDefinitionEnumeration),
        aslist=c.MixedElementList,
    )


@c.xtype_handler(None)
class ModuleType(AbstractType):
    """A requirement-module type."""

    _xmltag = "ownedTypes"


@c.xtype_handler(None)
class RelationType(AbstractType):
    """A requirement-relation type."""

    _xmltag = "ownedTypes"


@c.xtype_handler(None)
class RequirementType(AbstractType):
    """A requirement type."""

    _xmltag = "ownedTypes"


@c.xtype_handler(None)
class Requirement(ReqIFElement):
    """A ReqIF Requirement."""

    _xmltag = "ownedRequirements"

    owner = c.ParentAccessor(c.GenericElement)

    chapter_name = c.AttributeProperty("ReqIFChapterName", optional=True)
    foreign_id = c.AttributeProperty(
        "ReqIFForeignID", optional=True, returntype=int
    )
    text = c.HTMLAttributeProperty("ReqIFText", optional=True)
    attributes = AttributeAccessor()
    type = c.AttrProxyAccessor(RequirementType, "requirementType")

    relations: c.Accessor[AbstractRequirementsRelation]
    related: c.Accessor[c.GenericElement]


@c.xtype_handler(None)
class Folder(Requirement):
    """A folder that stores Requirements."""

    _xmltag = "ownedRequirements"

    folders: c.Accessor
    requirements = c.DirectProxyAccessor(Requirement, aslist=c.ElementList)


class AbstractRequirementsRelation(ReqIFElement):
    _required_attrs = frozenset({"source", "target"})

    type = c.AttrProxyAccessor(RelationType, "relationType")
    source = c.AttrProxyAccessor(Requirement, "source")
    target = c.AttrProxyAccessor(c.GenericElement, "target")

    def _short_repr_(self) -> str:
        direction = ""
        if self.source is not None:
            direction += f" from {self.source._short_repr_()}"
        if self.target is not None:
            direction += f" to {self.target._short_repr_()}"
        return (
            f"<{type(self).__name__} {self.long_name!r}{direction} "
            f"({self.uuid})>"
        )


@c.xtype_handler(None)
class InternalRelation(AbstractRequirementsRelation):
    """A Relation between two requirements."""

    _xmltag = "ownedRelations"


_attr_type_hints = {
    "int": IntegerValueAttribute,
    "integer": IntegerValueAttribute,
    "integervalueattribute": IntegerValueAttribute,
    "str": StringValueAttribute,
    "string": StringValueAttribute,
    "stringvalueattribute": StringValueAttribute,
    "float": RealValueAttribute,
    "real": RealValueAttribute,
    "realvalueattribute": RealValueAttribute,
    "date": DateValueAttribute,
    "datetime": DateValueAttribute,
    "datevalueattribute": DateValueAttribute,
    "bool": BooleanValueAttribute,
    "boolean": BooleanValueAttribute,
    "booleanvalueattribute": BooleanValueAttribute,
    "enum": EnumerationValueAttribute,
    "enumeration": EnumerationValueAttribute,
    "enumerationvalueattribute": EnumerationValueAttribute,
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Two dimensional vector calculation utility."""
# pylint: disable=unsubscriptable-object, not-an-iterable  # false-positives
from __future__ import annotations

__all__ = [
    "Vec2Element",
    "Vec2List",
    "Vec2Property",
    "Vec2ish",
    "Vector2D",
    "line_intersect",
]

import collections.abc as cabc
import math
import operator
import typing as t

Vec2Element = t.Union[float, int]
Vec2ish = t.Tuple[Vec2Element, Vec2Element]


class Vector2D(t.NamedTuple):
    """A vector in 2-dimensional space."""

    x: Vec2Element = 0
    y: Vec2Element = 0

    def __add__(self, other: Vec2ish) -> Vector2D:  # type: ignore[override]
        return self.__map2(operator.add, other)

    def __radd__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.add, other, True)

    def __sub__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.sub, other)

    def __rsub__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.sub, other, True)

    @t.overload  # type: ignore
    def __mul__(self, other: Vec2ish) -> Vec2Element: ...

    @t.overload
    def __mul__(self, other: Vec2Element) -> Vector2D: ...

    def __mul__(self, other: Vec2Element | Vec2ish) -> Vector2D | Vec2Element:
        result = self.__map2(operator.mul, other)
        if result is NotImplemented:
            return self.__map(operator.mul, other)
        return sum(result)

    @t.overload  # type: ignore[override]
    def __rmul__(self, other: Vec2ish) -> Vec2Element: ...

    @t.overload
    def __rmul__(self, other: Vec2Element) -> Vector2D: ...

    def __rmul__(self, other: Vec2Element | Vec2ish) -> Vector2D | Vec2Element:
        result = self.__map2(operator.mul, other, True)
        if result is NotImplemented:
            return self.__map(operator.mul, other, True)
        return sum(result)

    def __matmul__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.mul, other)

    def __rmatmul__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.mul, other, True)

    def __truediv__(self, other: Vec2Element) -> Vector2D:
        return self.__map(operator.truediv, other)

    def __rtruediv__(self, other: Vec2Element) -> Vector2D:
        return self.__map(operator.truediv, other, True)

    def __floordiv__(self, other: Vec2Element) -> Vector2D:
        result = self.__map(operator.floordiv, other)
        if result is NotImplemented:  # pragma: no cover
            return result
        return type(self)(int(result.x), int(result.y))

    def __rfloordiv__(self, other: Vec2Element) -> Vector2D:
        result = self.__map(operator.floordiv, other, True)
        if result is NotImplemented:  # pragma: no cover
            return result
        return type(self)(int(result.x), int(result.y))

    def __abs__(self) -> Vector2D:
        return type(self)(abs(self[0]), abs(self[1]))

    def __str__(self) -> str:  # pragma: no cover
        return f"({self[0]}, {self[1]})"

    @property
    def sqlength(self) -> float:
        """Calculate the squared length of this vector."""
        return self[0] ** 2 + self[1] ** 2

    @property
    def length(self) -> float:
        """Calculate the length of this vector."""
        return math.sqrt(self.sqlength)

    @property
    def normalized(self) -> Vector2D:
        """Create a unit Vector2D with the same direction as this one.

        Raises
        ------
        ZeroDivisionError
            if this Vector2D has zero length
        """
        length = self.length
        return Vector2D(*(i / length for i in self))

    def closestaxis(self) -> Vector2D:
        """Determine the axis closest to this Vector2D."""
        horizontal = abs(self[0]) >= abs(self[1])
        return Vector2D(*((-1, 1)[i >= 0] for i in self)) @ (
            horizontal,
            not horizontal,
        )

    def angleto(self, other: Vec2ish) -> float:
        r"""Calculate the angle to ``other``.

        This method calculates the angle that ``other`` was rotated by
        in order to have the same direction as ``self``, in radians.

        Notes
        -----
        The returned angle will always constitute the shortest rotation
        possible, i.e. it can have values between :math:`-\pi` and
        :math:`+\pi`.
        """
        angle = math.atan2(self[1], self[0]) - math.atan2(other[1], other[0])
        if angle > math.pi:  # pragma: no cover
            angle -= 2 * math.pi
        elif angle < -math.pi:  # pragma: no cover
            angle += 2 * math.pi
        return angle

    def rotatedby(self, theta: Vec2Element) -> Vector2D:
        """Rotate this Vector2D by ``theta`` radians."""
        cos_t = math.cos(theta)
        sin_t = math.sin(theta)

        return Vector2D(
            self[0] * cos_t - self[1] * sin_t,
            self[0] * sin_t + self[1] * cos_t,
        )

    def boxsnap(self, corner1: Vec2ish, corner2: Vec2ish) -> Vector2D:
        """Snap this vector to the side of a box and return the result.

        Parameters
        ----------
        corner1
            A Vector2D describing the first corner of the target box.
        corner2
            A Vector2D describing the second corner of the target box.
        dirvec
            Ignored.
        """
        minx = min(corner1[0], corner2[0])
        miny = min(corner1[1], corner2[1])
        maxx = max(corner1[0], corner2[0])
        maxy = max(corner1[1], corner2[1])
        topleft = Vector2D(minx, miny)
        bottomright = Vector2D(maxx, maxy)
        return self.__dirless_boxsnap(topleft, bottomright)

    def __dirless_boxsnap(
        self, topleft: Vector2D, bottomright: Vector2D
    ) -> Vector2D:
        x, y = self.x, self.y
        if x < topleft.x:
            x = topleft.x
        elif x > bottomright.x:
            x = bottomright.x
        if y < topleft.y:
            y = topleft.y
        elif y > bottomright.y:
            y = bottomright.y

        if self != (x, y):
            return Vector2D(x, y)

        distances = [
            Vector2D(topleft.x - self.x, 0),
            Vector2D(bottomright.x - self.x, 0),
            Vector2D(0, topleft.y - self.y),
            Vector2D(0, bottomright.y - self.y),
        ]
        offset = min(distances, key=lambda i: i.sqlength)
        return self + offset

    def __map(
        self,
        func: cabc.Callable[[Vec2Element, Vec2Element], Vec2Element],
        other: Vec2Element | Vec2ish,
        reflected: bool = False,
    ) -> Vector2D:
        if not isinstance(other, (int, float)):  # pragma: no cover
            return NotImplemented
        if reflected:
            return type(self)(func(other, self[0]), func(other, self[1]))
        return type(self)(func(self[0], other), func(self[1], other))

    def __map2(
        self,
        func: cabc.Callable[[Vec2Element, Vec2Element], Vec2Element],
        other: Vec2Element | Vec2ish,
        reflected: bool = False,
    ) -> Vector2D:
        if isinstance(other, (int, float)):  # pragma: no cover
            return NotImplemented
        if not len(other) == 2:  # pragma: no cover
            raise ValueError("Length of 'other' must be 2")
        if reflected:
            return type(self)(func(other[0], self[0]), func(other[1], self[1]))
        return type(self)(func(self[0], other[0]), func(self[1], other[1]))


class Vec2Property:
    """A property that automatically converts 2-tuples into Vector2D."""

    __slots__ = ("default", "name", "__objclass__")
    default: Vector2D | None
    name: str | None
    __objclass__: type[t.Any]

    def __init__(self, default: Vec2ish | None = None):
        if default is None or isinstance(default, Vector2D):
            self.default = default
        else:
            self.default = Vector2D(*default)

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> Vec2Property: ...

    @t.overload
    def __get__(
        self, obj: t.Any, objtype: type[t.Any] | None = ...
    ) -> Vector2D: ...

    def __get__(
        self, obj: t.Any | None, objtype: type[t.Any] | None = None
    ) -> Vec2Property | Vector2D:
        if obj is None:
            return self
        if self.name is None:
            raise RuntimeError("This property does not have a name yet")
        try:
            return getattr(obj, f"_{type(self).__name__}__{self.name}")
        except AttributeError:
            if self.default:
                return self.default
            raise

    def __set__(self, obj: t.Any, value: Vec2ish) -> None:
        if self.name is None:
            raise RuntimeError("This property does not have a name yet")
        if not isinstance(value, Vector2D):
            value = Vector2D(*value)
        setattr(obj, f"_{type(self).__name__}__{self.name}", value)

    def __set_name__(self, owner: type[t.Any], name: str) -> None:
        self.__objclass__ = owner
        self.name = name


class Vec2List(t.MutableSequence[Vector2D]):
    """A list that automatically converts its elements into Vector2D."""

    def __init__(self, values: cabc.Iterable[Vec2ish]):
        self.__list: list[Vector2D] = []
        self.extend(values)

    def __len__(self) -> int:
        return len(self.__list)

    @t.overload
    def __getitem__(self, index: int) -> Vector2D: ...

    @t.overload
    def __getitem__(self, index: slice) -> cabc.MutableSequence[Vector2D]: ...

    def __getitem__(
        self, index: int | slice
    ) -> Vector2D | cabc.Sequence[Vector2D]:
        return self.__list[index]

    @t.overload
    def __setitem__(
        self,
        index: int,
        value: Vec2ish,
    ) -> None: ...

    @t.overload
    def __setitem__(
        self,
        index: slice,
        value: cabc.Iterable[Vec2ish],
    ) -> None: ...

    def __setitem__(
        self,
        index: int | slice,
        value: Vec2ish | cabc.Iterable[Vec2ish],
    ) -> None:
        if isinstance(index, slice):
            assert not isinstance(value, Vector2D)
            value = t.cast(cabc.Iterable[Vec2ish], value)
            self.__list[index] = (self.__cast(v) for v in value)
        else:
            assert isinstance(value, Vector2D)
            self.__list[index] = self.__cast(value)

    def __delitem__(self, index: int | slice) -> None:
        del self.__list[index]

    def append(self, value: Vec2ish) -> None:
        self.__list.append(self.__cast(value))

    def copy(self) -> Vec2List:
        """Create a copy of this Vec2List."""
        return Vec2List(self)

    def extend(self, values: cabc.Iterable[Vec2ish]) -> None:
        for i in values:
            self.append(i)

    def insert(self, index: int, value: Vector2D) -> None:
        self.__list.insert(index, self.__cast(value))

    @staticmethod
    def __cast(element: Vec2ish) -> Vector2D:
        if not isinstance(element, Vector2D):
            element = Vector2D(*element)
        return element


def line_intersect(
    line1: tuple[Vec2ish, Vec2ish], line2: tuple[Vec2ish, Vec2ish]
) -> Vector2D:
    """Calculate the point where ``line1`` and ``line2`` intersect.

    Both lines are straight lines with infinite length that are defined
    by the given points.

    Notes
    -----
    The implementation is based on
    `https://mathworld.wolfram.com/Line-LineIntersection.html`.
    """
    (x1, y1), (x2, y2) = line1
    (x3, y3), (x4, y4) = line2
    denum = (x1 - x2) * (y3 - y4) - (x3 - x4) * (y1 - y2)
    if denum == 0:
        raise ValueError("Lines are parallel")

    d1 = x1 * y2 - x2 * y1
    d2 = x3 * y4 - x4 * y3
    x = d1 * (x3 - x4) - d2 * (x1 - x2)
    y = d1 * (y3 - y4) - d2 * (y1 - y2)
    return Vector2D(x / denum, y / denum)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Abstract classes acting as templates for concrete classes.

These base classes are used between different layers.
"""

from .. import common as c


class TraceableElement(c.GenericElement):
    """A template for traceable ModelObjects."""

    source = c.AttrProxyAccessor(c.GenericElement, attr="sourceElement")
    target = c.AttrProxyAccessor(c.GenericElement, attr="targetElement")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=abstract-method, useless-suppression
# For some reason, pylint in Github CI didn't get the memo that these aren't
# actually abstract methods. Other pylint installations seem to agree that
# implementing these methods isn't necessary. So we just ignore the warning
# about that here.
# TODO Revisit this decision some time in the future

from __future__ import annotations

__all__ = [
    "MemoryFileHandler",
    "MemoryFilePath",
    "MemoryFile",
]

import collections.abc as cabc
import io
import os
import pathlib
import typing as t

from capellambse import helpers

from . import abc

if t.TYPE_CHECKING:
    from capellambse.loader.modelinfo import ModelInfo


class MemoryFileHandler(abc.FileHandler):
    """A file handler that stores data in memory."""

    def __init__(
        self,
        path: str | os.PathLike = "memory:",
        *,
        subdir: str | pathlib.PurePosixPath = "/",
    ) -> None:
        """Initialize a new memory file handler.

        Parameters
        ----------
        path
            An optional path to a directory to use as fallback. Opened
            files' contents will be prepopulated with the contents of
            files from this directory.
        subdir
            An optional path to prepend to all opened (physical) files.
        """
        if path != "memory:":
            raise ValueError(f"Unsupported path for MemoryFileHandler: {path}")
        super().__init__(path, subdir=subdir)

        self._data: dict[pathlib.PurePosixPath, bytearray] = {}

    def get_model_info(self) -> ModelInfo:
        return ModelInfo(url="memory:")

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        path = helpers.normalize_pure_path(filename, base=self.subdir)
        if "w" in mode:
            self._data[path] = bytearray()
            return MemoryFile(self._data[path], "w")  # type: ignore[abstract]

        try:
            return MemoryFile(self._data[path], "r")  # type: ignore[abstract]
        except KeyError:
            pass
        raise FileNotFoundError(path)

    def __repr__(self) -> str:
        return (
            f"<memory with {len(self._data)} files: "
            f"{', '.join(map(str, self._data.keys()))}>"
        )

    @property
    def rootdir(self) -> MemoryFilePath:
        return MemoryFilePath(self, pathlib.PurePosixPath("."))

    def iterdir(
        self, path: str | pathlib.PurePosixPath = "/", /
    ) -> cabc.Iterator[MemoryFilePath]:
        path = helpers.normalize_pure_path(path)
        for p in self._data:
            if p.parent == path:
                yield MemoryFilePath(self, p)


class MemoryFile(t.BinaryIO):
    def __init__(self, data: bytearray, mode: t.Literal["r", "w"]) -> None:
        self._data = data
        self._mode = mode
        self._pos = 0

    def __enter__(self) -> MemoryFile:
        return self

    def __exit__(self, *args: t.Any) -> None:
        pass

    def write(self, s: bytes | bytearray) -> int:  # type: ignore[override]
        if self._mode != "w":
            raise io.UnsupportedOperation("not writable")
        self._data[self._pos : self._pos + len(s)] = s
        self._pos += len(s)
        return len(s)

    def read(self, n: int = -1) -> bytes:
        if self._mode != "r":
            raise io.UnsupportedOperation("not readable")
        if n < 0:
            n = len(self._data) - self._pos
        result = self._data[self._pos : self._pos + n]
        self._pos += len(result)
        return bytes(result)


class MemoryFilePath(abc.AbstractFilePath[MemoryFileHandler]):
    def is_dir(self) -> bool:
        return not self.is_file()

    def is_file(self) -> bool:
        return self._path in self._parent._data



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for the System Analysis layer.

This is normally the place to declare data used in the model for e.g.
functions, actors etc. which is best presented in a glossary document.

.. diagram:: [CDB] SA ORM
"""
from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import capellacommon, capellacore, cs, fa, interaction
from . import oa

XT_ARCH = "org.polarsys.capella.core.data.ctx:SystemAnalysis"


@c.xtype_handler(XT_ARCH)
class SystemFunction(fa.Function):
    """A system function."""

    _xmltag = "ownedFunctions"

    realized_operational_activities = c.TypecastAccessor(
        oa.OperationalActivity, "realized_functions"
    )

    owner: c.Accessor


@c.xtype_handler(XT_ARCH)
class SystemFunctionPkg(c.GenericElement):
    """A function package that can hold functions."""

    _xmltag = "ownedFunctionPkg"

    functions = c.DirectProxyAccessor(SystemFunction, aslist=c.ElementList)
    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class SystemComponent(cs.Component):
    """A system component."""

    _xmltag = "ownedSystemComponents"

    allocated_functions = c.LinkAccessor[SystemFunction](
        "ownedFunctionalAllocation",
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        attr="targetElement",
    )
    realized_entities = c.TypecastAccessor(
        oa.Entity,
        "realized_components",
    )
    realized_operational_entities = c.TypecastAccessor(
        oa.Entity,
        "realized_components",
    )


@c.xtype_handler(XT_ARCH)
class SystemComponentPkg(c.GenericElement):
    """A system component package."""

    _xmltag = "ownedSystemComponentPkg"

    components = c.DirectProxyAccessor(SystemComponent, aslist=c.ElementList)
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityInvolvement(interaction.AbstractInvolvement):
    """A CapabilityInvolvement."""


@c.xtype_handler(XT_ARCH)
class Capability(c.GenericElement):
    """A capability."""

    _xmltag = "ownedCapabilities"

    extends = c.DirectProxyAccessor(
        interaction.AbstractCapabilityExtend, aslist=c.ElementList
    )
    extended_by = c.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityExtend, "target", aslist=c.ElementList
    )
    includes = c.DirectProxyAccessor(
        interaction.AbstractCapabilityInclude, aslist=c.ElementList
    )
    included_by = c.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityInclude, "target", aslist=c.ElementList
    )
    generalizes = c.DirectProxyAccessor(
        interaction.AbstractCapabilityGeneralization, aslist=c.ElementList
    )
    generalized_by = c.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityGeneralization,
        "target",
        aslist=c.ElementList,
    )
    owned_chains = c.DirectProxyAccessor(
        fa.FunctionalChain, aslist=c.ElementList
    )
    involved_functions = c.LinkAccessor[SystemFunction](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        interaction.AbstractFunctionAbstractCapabilityInvolvement,
        aslist=c.ElementList,
        attr="involved",
    )
    involved_chains = c.LinkAccessor[fa.FunctionalChain](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        interaction.XT_CAP2PROC,
        aslist=c.ElementList,
        attr="involved",
    )
    involved_components = c.LinkAccessor[SystemComponent](
        "ownedCapabilityInvolvements",
        CapabilityInvolvement,
        aslist=c.MixedElementList,
        attr="involved",
    )
    component_involvements = c.DirectProxyAccessor(
        CapabilityInvolvement, aslist=c.ElementList
    )
    realized_capabilities = c.LinkAccessor[oa.OperationalCapability](
        None,  # FIXME fill in tag
        interaction.XT_CAP_REAL,
        aslist=c.ElementList,
        attr="targetElement",
    )

    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = c.DirectProxyAccessor(
        interaction.Scenario, aslist=c.ElementList
    )
    states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class MissionInvolvement(interaction.AbstractInvolvement):
    """A MissionInvolvement."""

    _xmltag = "ownedMissionInvolvements"


@c.xtype_handler(XT_ARCH)
class CapabilityExploitation(c.GenericElement):
    """A CapabilityExploitation."""

    _xmltag = "ownedCapabilityExploitations"

    capability = c.AttrProxyAccessor(Capability, "capability")

    @property
    def name(self) -> str:  # type: ignore[override]
        """Return the name."""
        direction = ""
        if self.capability is not None:
            direction = f" to {self.capability.name} ({self.capability.uuid})"

        return f"[{self.__class__.__name__}]{direction}"


@c.xtype_handler(XT_ARCH)
class Mission(c.GenericElement):
    """A mission."""

    _xmltag = "ownedMissions"

    involvements = c.DirectProxyAccessor(
        MissionInvolvement, aslist=c.ElementList
    )
    incoming_involvements = c.ReferenceSearchingAccessor(
        MissionInvolvement, "target", aslist=c.ElementList
    )
    exploits = c.LinkAccessor[Capability](
        None,  # FIXME fill in tag
        CapabilityExploitation,
        aslist=c.ElementList,
        attr="capability",
    )
    exploitations = c.DirectProxyAccessor(
        CapabilityExploitation, aslist=c.ElementList
    )


@c.xtype_handler(XT_ARCH)
class MissionPkg(c.GenericElement):
    """A system mission package that can hold missions."""

    _xmltag = "ownedMissionPkg"

    missions = c.DirectProxyAccessor(Mission, aslist=c.ElementList)
    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityPkg(c.GenericElement):
    """A capability package that can hold capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = c.DirectProxyAccessor(Capability, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(None)
class SystemAnalysis(crosslayer.BaseArchitectureLayer):
    """Provides access to the SystemAnalysis layer of the model."""

    root_component = c.AttributeMatcherAccessor(
        SystemComponent,
        attributes={"is_actor": False},
        rootelem=SystemComponentPkg,
    )
    root_function = c.DirectProxyAccessor(
        SystemFunction, rootelem=SystemFunctionPkg
    )

    function_package = c.DirectProxyAccessor(SystemFunctionPkg)
    capability_package = c.DirectProxyAccessor(CapabilityPkg)
    component_package = c.DirectProxyAccessor(SystemComponentPkg)
    mission_package = c.DirectProxyAccessor(MissionPkg)

    all_functions = c.DeepProxyAccessor(SystemFunction, aslist=c.ElementList)
    all_capabilities = c.DeepProxyAccessor(Capability, aslist=c.ElementList)
    all_components = c.DeepProxyAccessor(SystemComponent, aslist=c.ElementList)
    all_actors = property(
        lambda self: self._model.search(SystemComponent).by_is_actor(True)
    )
    all_missions = c.DeepProxyAccessor(Mission, aslist=c.ElementList)
    all_functional_chains = property(
        lambda self: self._model.search(fa.FunctionalChain, below=self)
    )

    actor_exchanges = c.DirectProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=SystemComponentPkg,
    )
    component_exchanges = c.DeepProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=[SystemComponentPkg, SystemComponent],
    )

    all_capability_exploitations = c.DeepProxyAccessor(
        CapabilityExploitation, aslist=c.ElementList
    )
    all_function_exchanges = c.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[SystemFunctionPkg, SystemFunction],
    )
    all_component_exchanges = c.DeepProxyAccessor(
        fa.ComponentExchange, aslist=c.ElementList
    )

    diagrams = diagram.DiagramAccessor(
        "System Analysis", cacheattr="_MelodyModel__diagram_cache"
    )


c.set_accessor(
    SystemFunction,
    "owner",
    c.ReferenceSearchingAccessor(SystemComponent, "allocated_functions"),
)
c.set_accessor(
    SystemFunction,
    "packages",
    c.DirectProxyAccessor(SystemFunctionPkg, aslist=c.ElementList),
)
c.set_accessor(
    oa.OperationalCapability,
    "realizing_capabilities",
    c.ReferenceSearchingAccessor(
        Capability, "realized_capabilities", aslist=c.ElementList
    ),
)
c.set_accessor(
    Capability,
    "incoming_exploitations",
    c.ReferenceSearchingAccessor(
        CapabilityExploitation, "capability", aslist=c.ElementList
    ),
)
c.set_accessor(
    oa.Entity,
    "realizing_system_components",
    c.ReferenceSearchingAccessor(
        SystemComponent, "realized_operational_entities", aslist=c.ElementList
    ),
)
c.set_accessor(
    oa.OperationalActivity,
    "realizing_system_functions",
    c.ReferenceSearchingAccessor(
        SystemFunction, "realized_operational_activities", aslist=c.ElementList
    ),
)
c.set_self_references(
    (MissionPkg, "packages"),
    (SystemComponent, "components"),
    (SystemComponentPkg, "packages"),
    (SystemFunction, "functions"),
    (SystemFunctionPkg, "packages"),
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0


from ... import common as c


class LiteralBooleanValue(c.GenericElement):
    """A Literal Boolean Value."""

    _xmltag = "ownedLiterals"

    value = c.BooleanAttributeProperty("value")


class LiteralValue(c.GenericElement):
    is_abstract = c.BooleanAttributeProperty(
        "abstract", __doc__="Indicates if property is abstract"
    )
    value = c.AttributeProperty("value", optional=True, returntype=str)
    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")


@c.xtype_handler(None)
class LiteralNumericValue(LiteralValue):
    value = c.AttributeProperty("value", optional=True)
    unit = c.AttrProxyAccessor(c.GenericElement, "unit")


@c.xtype_handler(None)
class LiteralStringValue(LiteralValue):
    """A Literal String Value."""


@c.xtype_handler(None)
class ValuePart(c.GenericElement):
    """A Value Part of a Complex Value."""

    _xmltag = "ownedParts"

    referenced_property = c.AttrProxyAccessor(
        c.GenericElement, "referencedProperty"
    )
    value = c.RoleTagAccessor("ownedValue")


@c.xtype_handler(None)
class ComplexValue(c.GenericElement):
    """A Complex Value."""

    _xmltag = "ownedDataValues"

    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    value_parts = c.DirectProxyAccessor(ValuePart, aslist=c.ElementList)


@c.attr_equal("name")
@c.xtype_handler(None)
class EnumerationLiteral(c.GenericElement):
    """An EnumerationLiteral (proxy link)."""

    _xmltag = "ownedLiterals"

    name = c.AttributeProperty("name", returntype=str)
    value = c.RoleTagAccessor("domainValue")

    owner: c.Accessor

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.name == other
        return super().__eq__(other)


@c.xtype_handler(None)
class EnumerationReference(c.GenericElement):
    name = c.AttributeProperty("name", returntype=str)
    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    value = c.AttrProxyAccessor(c.GenericElement, "referencedValue")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Miscellaneous utility functions used throughout the modules."""
from __future__ import annotations

import collections
import collections.abc as cabc
import contextlib
import errno
import functools
import html
import importlib.resources as imr
import itertools
import logging
import math
import operator
import pathlib
import re
import sys
import time
import typing as t

import lxml.html
import markupsafe
import typing_extensions as te
from lxml import etree
from PIL import ImageFont

import capellambse
import capellambse._namespaces as _n

if sys.platform.startswith("win"):
    import msvcrt
else:
    import fcntl

LOGGER = logging.getLogger(__name__)

ATT_XT = f"{{{_n.NAMESPACES['xsi']}}}type"
FALLBACK_FONT = "OpenSans-Regular.ttf"
RE_TAG_NS = re.compile(r"(?:\{(?P<ns>[^}]*)\})?(?P<tag>.*)")
RE_VALID_UUID = re.compile(
    r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}",
    re.IGNORECASE,
)
LINEBREAK_AFTER = frozenset({"br", "p", "ul", "li"})
TABS_BEFORE = frozenset({"li"})

CROSS_FRAGMENT_LINK = re.compile(
    r"""
    ^
    (?:
        (?:
            (?:(?P<xtype>[^ #]+)\ )?
            (?P<fragment>[^ #]+)
        )?
        \#
    )?
    (?P<uuid>[A-Za-z0-9_-]+)
    $
    """,
    re.VERBOSE,
)

UUIDString = t.NewType("UUIDString", str)
_T = t.TypeVar("_T")


def flatten_html_string(text: str) -> str:
    """Convert an HTML-string to plain text."""
    frags = lxml.html.fragments_fromstring(text)
    if not frags:
        return ""

    text_container: list[str] = []
    if isinstance(frags[0], str):
        text_container.append(frags.pop(0))

    for frag in frags:
        text_container.extend(_flatten_subtree(frag))

    return "".join(text_container).rstrip()


def _flatten_subtree(element: etree._Element) -> cabc.Iterator[str]:
    def remove_whitespace(text: str):
        return re.sub("[\n\t]", "", text).lstrip()

    if element.tag in TABS_BEFORE:
        yield "             • "

    if element.text:
        yield remove_whitespace(element.text)

    for child in element:
        yield from _flatten_subtree(child)

    if element.tag in LINEBREAK_AFTER:
        yield "\n"

    if element.tail:
        yield remove_whitespace(element.tail)


def is_uuid_string(string: t.Any) -> te.TypeGuard[UUIDString]:
    """Validate that ``string`` is a valid UUID."""
    return isinstance(string, str) and bool(RE_VALID_UUID.fullmatch(string))


# File name and path manipulation
def normalize_pure_path(
    path: str | pathlib.PurePosixPath,
    *,
    base: str | pathlib.PurePosixPath = "/",
) -> pathlib.PurePosixPath:
    """Make a PurePosixPath relative to *base* and collapse ``..`` components.

    Parameters
    ----------
    path
        The input path to normalize.
    base
        The base directory to which relative paths should be
        interpreted. Ignored if the input path is not relative.

    Returns
    -------
    pathlib.PurePosixPath
        The normalized path.
    """
    path = pathlib.PurePosixPath("/", base, path)

    parts: list[str] = []
    for i in path.parts[1:]:
        if i == "..":
            try:
                parts.pop()
            except IndexError:
                pass
        else:
            parts.append(i)

    return pathlib.PurePosixPath(*parts)


def relpath_pure(
    path: pathlib.PurePosixPath, start: pathlib.PurePosixPath
) -> pathlib.PurePosixPath:
    """Calculate the relative path between two pure paths.

    Unlike :meth:`pathlib.PurePath.relative_to`, this method can cope
    with ``path`` not being a subpath of ``start``. And unlike the
    :func:`os.path.relpath` function, it does not involve any filesystem
    access.
    """
    parts = list(reversed(path.parts))
    prefix = True
    for part in start.parts:
        if prefix:
            if parts and parts[-1] == part:
                parts.pop()
            else:
                prefix = False
        else:
            parts.append("..")
    return pathlib.PurePosixPath(*reversed(parts))


if sys.platform.startswith("win"):

    @contextlib.contextmanager
    def flock(file: pathlib.Path) -> t.Iterator[None]:
        file = file.resolve()
        logged = False
        with file.open("wb") as lock:
            while True:
                try:
                    msvcrt.locking(lock.fileno(), msvcrt.LK_LOCK, 1)
                except OSError as err:
                    if err.errno == errno.EDEADLOCK:
                        if not logged:
                            LOGGER.debug("Waiting for lock file %s", file)
                            logged = True
                        time.sleep(1)
                    else:
                        raise
                else:
                    break

            try:
                yield
            finally:
                msvcrt.locking(lock.fileno(), msvcrt.LK_UNLCK, 1)

else:

    @contextlib.contextmanager
    def flock(file: pathlib.Path) -> t.Iterator[None]:
        file = file.resolve()
        with file.open("wb") as lock:
            try:
                fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except BlockingIOError:
                LOGGER.debug("Waiting for lock file %s", file)
                fcntl.flock(lock, fcntl.LOCK_EX)

            yield


# Text processing and rendering
@functools.lru_cache(maxsize=8)
def load_font(fonttype: str, size: int) -> ImageFont.FreeTypeFont:
    for name in (fonttype, fonttype.upper(), fonttype.lower()):
        try:
            return ImageFont.truetype(name, size)
        except OSError:
            pass

    fontfile = imr.files(capellambse).joinpath(FALLBACK_FONT)
    with fontfile.open("rb") as fallback_font:
        return ImageFont.truetype(fallback_font, size)


@functools.lru_cache(maxsize=256)
def extent_func(
    text: str,
    fonttype: str = "segoeui.ttf",
    size: int = 8,
) -> tuple[float, float]:
    """Calculate the display size of the given text.

    Parameters
    ----------
    text
        Text to calculate pixel size on
    fonttype
        The font type / face
    size
        Font size (px)

    Returns
    -------
    width
        The calculated width of the text (px).
    height
        The calculated height of the text (px).
    """
    width = height = 0
    font = load_font(fonttype, size)
    (width, height), _ = font.font.getsize(text)
    return (width * 10 / 7, height * 10 / 7)


def get_text_extent(
    text: str,
    width: float | int = math.inf,
) -> tuple[float, float]:
    """Calculate the bounding box size of ``text`` after line wrapping.

    Parameters
    ----------
    text
        Text to calculate the size for.
    width
        Maximum line length (px).

    Returns
    -------
    width
        The width of the text after word wrapping (px).
    height
        The height of the text after word wrapping (px).
    """
    lines = [*map(extent_func, word_wrap(text, width))]
    line_height = max(l[1] for l in lines)
    return max(l[0] for l in lines), line_height * len(lines)


def ssvparse(
    string: str,
    cast: cabc.Callable[[str], _T],
    *,
    parens: cabc.Sequence[str] = ("", ""),
    sep: str = ",",
    num: int = 0,
) -> list[_T]:
    """Parse a string of ``sep``-separated values wrapped in ``parens``.

    Parameters
    ----------
    string
        The input string.
    cast
        A type to cast the values into.
    parens
        The parentheses that must exist around the input. Either a
        two-character string or a 2-tuple of strings.
    sep
        The separator between values.
    num
        If non-zero, only accept exactly this many values.

    Returns
    -------
    list[_T]
        A list of values cast into the given type.

    Raises
    ------
    ValueError
        If the parentheses are missing around the input string, or if
        the expected number of values doesn't match the actual number.
    """
    if not string.startswith(parens[0]) or not string.endswith(parens[1]):
        raise ValueError(f"Missing {parens} around string: {string}")
    string = string[len(parens[0]) : -len(parens[1])]
    values = [cast(v) for v in string.split(sep)]
    if num and len(values) != num:
        raise ValueError(
            f"Expected {num} values, found {len(values)}: {string}"
        )
    return values


def word_wrap(text: str, width: float | int) -> list[str]:
    """Perform word wrapping for proportional fonts.

    Whitespace at the beginning of input lines is preserved, but other
    whitespace is collapsed to single spaces.

    Parameters
    ----------
    text
        The text to wrap.
    width
        The width in pixels to wrap to.

    Returns
    -------
    list[str]
        A list of strings, one for each line, after wrapping.
    """

    def rejoin(words: cabc.Iterable[str], start: int, stop: int | None) -> str:
        return " ".join(itertools.islice(words, start, stop))

    def splitline(line: str) -> list[str]:
        match = re.search(r"^\s*", line)
        assert match is not None
        words = line.split()

        if words:
            words[0] = match.group(0) + words[0]
        return words

    output_lines = []
    input_lines = collections.deque(text.splitlines())
    while input_lines:
        words = collections.deque(splitline(input_lines.popleft()))
        if not words:
            output_lines.append("")
            continue

        words_count = len(words)
        while (
            extent_func(rejoin(words, 0, words_count))[0] > width
            and words_count > 0
        ):
            words_count -= 1

        if words_count > 0:
            output_lines.append(rejoin(words, 0, words_count))
            if words_count < len(words):
                input_lines.appendleft(rejoin(words, words_count, None))

        else:
            word = words.popleft()
            letters_count = len(word)
            while (
                extent_func(word[:letters_count])[0] > width
                and letters_count > 1
            ):
                letters_count -= 1

            output_lines.append(word[:letters_count])
            if letters_count < len(word):
                words.appendleft(word[letters_count:])

            input_lines.appendleft(" ".join(words))

    return output_lines or [""]


# XML tree modification and navigation
def repair_html(markup: str) -> markupsafe.Markup:
    """Try to repair broken HTML markup to prevent parse errors.

    Parameters
    ----------
    markup
        The markup to try and repair.

    Returns
    -------
    markupsafe.Markup
        The repaired markup.
    """

    def cb(node: etree._Element) -> None:
        for k in list(node.keys()):
            if ":" in k:
                del node.attrib[k]

    return process_html_fragments(markup, cb)


def process_html_fragments(
    markup: str, node_callback: cabc.Callable[[etree._Element], None]
) -> markupsafe.Markup:
    """Repair and modify HTML markup.

    The original markup, which can be an HTML fragment (without a root
    element), is parsed and processed, and then reassembled into a
    Markup instance. If the original markup contained any errors or
    inconsistencies, these are repaired in the returned Markup instance.

    Parameters
    ----------
    markup
        The markup string to modify.
    node_callback
        A callback function to process each node in the parsed markup.
        The function should accept a single
        :external:py:class:`lxml.etree._Element` as argument; its return
        value is ignored.

        Note that, since the markup is parsed as fragments, more than
        the first element passed to the callback may have no parent.

        The callback will not be invoked for leading text, if there is
        any, and thus it has no ability to influence it.

    Returns
    -------
    markupsafe.Markup
        The processed markup.
    """
    nodes: list[str | lxml.html._Element]
    nodes = lxml.html.fragments_fromstring(markup)
    if nodes and isinstance(nodes[0], str):
        firstnode: str = html.escape(nodes.pop(0))
    else:
        firstnode = ""
    assert all(isinstance(i, etree._Element) for i in nodes)

    for node in itertools.chain.from_iterable(
        map(operator.methodcaller("iter"), nodes)
    ):
        node_callback(node)

    othernodes = b"".join(
        etree.tostring(i, encoding="utf-8") for i in nodes
    ).decode("utf-8")

    return markupsafe.Markup(firstnode + othernodes)


def resolve_namespace(tag: str) -> str:
    """Resolve a ':'-delimited symbolic namespace to its canonical form.

    Parameters
    ----------
    tag
        Symbolic namespace delimited by ':'.

    Returns
    -------
    str
        Tag string in canonical form.
    """
    if ":" in tag:
        namespace, tag = tag.split(":")
        return f"{{{_n.NAMESPACES[namespace]}}}{tag}"
    return tag


def unescape_linked_text(
    loader: capellambse.loader.MelodyLoader, attr_text: str | None
) -> markupsafe.Markup:
    """Transform the ``linkedText`` into regular HTML."""

    def flatten_element(
        elm: str | lxml.html.HTMLElement,
    ) -> cabc.Iterator[str]:
        if isinstance(elm, str):
            yield html.escape(elm)
        elif elm.tag == "a":
            href = elm.get("href")
            if href is None:
                yield "&lt;broken link&gt;"
                yield html.escape(elm.tail or "")
                return
            ehref = html.escape(href)

            try:
                target = loader[href]
            except KeyError:
                yield f"&lt;deleted element {ehref}&gt;"
            else:
                if name := target.get("name"):
                    name = html.escape(name)
                else:
                    name = f"&lt;unnamed element {ehref}&gt;"
                yield f'<a href="hlink://{ehref}">{name}</a>'
            yield html.escape(elm.tail or "")
        else:
            yield html.escape(elm.text or "")
            for child in elm:
                yield from flatten_element(child)
            yield html.escape(elm.tail or "")

    elements = lxml.html.fragments_fromstring(attr_text or "")
    escaped_text = "".join(
        itertools.chain.from_iterable(flatten_element(i) for i in elements)
    )
    return markupsafe.Markup(escaped_text)


def escape_linked_text(
    loader: capellambse.loader.MelodyLoader, attr_text: str
) -> str:
    """Transform simple HTML with object links into ``LinkedText``.

    This is the inverse operation of :func:`unescape_linked_text`.
    """
    del loader

    def flatten_element(
        elm: str | lxml.html.HTMLElement,
    ) -> cabc.Iterator[str]:
        if isinstance(elm, str):
            yield html.escape(elm)
        elif elm.tag == "a":
            href = elm.get("href")
            if not (href or "").startswith("hlink://"):
                yield html.escape(elm.text or "")
            else:
                yield '<a href="'
                yield html.escape(href[len("hlink://") :])
                yield '"/>'
            if len(elm) > 0:
                raise ValueError("Nesting is not allowed in LinkedText")
        else:
            raise ValueError(
                f"Only 'a' tags are allowed in LinkedText, not {elm.tag!r}"
            )

    elements = lxml.html.fragments_fromstring(attr_text)
    text = "".join(
        itertools.chain.from_iterable(flatten_element(i) for i in elements)
    )
    return markupsafe.Markup(text)


def split_links(links: str) -> cabc.Iterator[str]:
    """Split a string containing intra- and inter-fragment links.

    Intra-fragment links are simply "#UUID", whereas inter-fragment
    links look like "xtype fragment#UUID". Multiple such links are
    space-separated in a single long string to form a list. This
    function splits such a string back into its individual components
    (each being either an intra- or inter-fragment link), and yields
    them.

    Yields
    ------
    str
        A single link from the list.
    """
    next_xtype = ""
    for part in links.split():
        if "#" in part:
            if next_xtype:
                part = f"{next_xtype} {part}"
                next_xtype = ""
            if not CROSS_FRAGMENT_LINK.fullmatch(part):
                raise ValueError(f"Malformed link definition: {links}")
            yield part

        else:
            if next_xtype:
                raise ValueError(f"Malformed link definition: {links}")
            next_xtype = part
    if next_xtype:
        raise ValueError(f"Malformed link definition: {links}")


@t.overload
def xpath_fetch_unique(
    xpath: str | etree.XPath,
    tree: etree._Element,
    elm_name: str,
    elm_uid: str | None = None,
    *,
    optional: t.Literal[False] = ...,
) -> etree._Element: ...


@t.overload
def xpath_fetch_unique(
    xpath: str | etree.XPath,
    tree: etree._Element,
    elm_name: str,
    elm_uid: str | None = None,
    *,
    optional: t.Literal[True],
) -> etree._Element | None: ...


def xpath_fetch_unique(
    xpath: str | etree.XPath,
    tree: etree._Element,
    elm_name: str,
    elm_uid: str | None = None,
    *,
    optional: bool = False,
) -> etree._Element | None:
    """Fetch an XPath result from the tree, ensuring that it's unique.

    Parameters
    ----------
    xpath
        The :class:`lxml.etree.XPath` object to apply, or an XPath
        expression as str.
    tree
        The (sub-)tree to which the XPath will be applied.
    elm_name
        A human-readable element name for error messages.
    elm_uid
        UID of the element which triggered this lookup. Will be included
        in the error message if an error occured.
    optional
        True to return None in case the element is not found. Otherwise
        a ValueError will be raised.

    Returns
    -------
    lxml.etree._Element | None
        The Element found by given ``xpath``.

    Raises
    ------
    ValueError
        If more than one element was found matching the ``xpath``, or if
        ``optional`` is ``False`` and no matching element was found.
    """
    if isinstance(xpath, str):
        xpath = etree.XPath(
            xpath, namespaces=_n.NAMESPACES, smart_strings=False
        )

    result = xpath(tree)
    if len(result) > 1:
        raise ValueError(
            f"Invalid XML: {elm_name!r} is not unique, found {len(result)}"
            + (f" while processing element {elm_uid!r}" if elm_uid else "")
        )
    if not optional and not result:
        raise ValueError(
            f"Invalid XML: {elm_name!r} not found"
            + (f" while processing element {elm_uid!r}" if elm_uid else "")
        )

    return result[0] if result else None


def xtype_of(elem: etree._Element) -> str | None:
    """Return the ``xsi:type`` of the element.

    If the element has an ``xsi:type`` attribute, its value is returned.

    If the element does not have an ``xsi:type``, this function resolves
    the tag's namespace to the symbolic name and reconstructs the type
    with the ``namespace:tag`` template.

    Parameters
    ----------
    elem
        The :class:`lxml.etree._Element` object to return the
        ``xsi:type`` for.

    Raises
    ------
    capellambse.UnsupportedPluginError
        If the plugin is unknown and therefore not supported.
    capellambse.UnsupportedPluginVersionError
        If the plugin's version is not supported.

    Returns
    -------
    str | None
        The ``xsi:type`` string of the provided element or ``None`` if
        the type could not be determined.
    """
    xtype = elem.get(ATT_XT)
    if xtype:
        return xtype

    tagmatch = RE_TAG_NS.fullmatch(elem.tag)
    assert tagmatch is not None
    ns = tagmatch.group("ns")
    tag = tagmatch.group("tag")
    if not ns:
        return None

    nskey, plugin = _n.get_keys_and_plugins_from_namespaces_by_url(ns)
    _n.check_plugin(nskey, plugin)
    return f"{nskey}:{tag}"


# More iteration tools
@t.overload
def ntuples(
    num: int, iterable: cabc.Iterable[_T], *, pad: t.Literal[False] = ...
) -> cabc.Iterator[tuple[_T, ...]]: ...


@t.overload
def ntuples(
    num: int, iterable: cabc.Iterable[_T], *, pad: t.Literal[True]
) -> cabc.Iterator[tuple[_T | None, ...]]: ...


def ntuples(
    num: int,
    iterable: cabc.Iterable[_T],
    *,
    pad: bool = False,
) -> cabc.Iterator[tuple[_T | None, ...]]:
    r"""Yield N items of ``iterable`` at once.

    Parameters
    ----------
    num
        The number of items to yield at once.
    iterable
        An iterable.
    pad
        If the items in ``iterable`` are not evenly divisible by ``n``,
        pad the last yielded tuple with ``None``\ s. If False, the last
        tuple will be discarded.

    Yields
    ------
    items
        A ``num`` long tuple of items from ``iterable``.
    """
    iterable = iter(iterable)
    while True:
        value = tuple(itertools.islice(iterable, num))
        if len(value) == num:
            yield value
        elif value and pad:
            yield value + (None,) * (num - len(value))
        else:
            break


# Simple one-trick helper classes
class EverythingContainer(t.Container[t.Any]):
    """A container that contains everything."""

    def __contains__(self, _: t.Any) -> bool:  # pragma: no cover
        """Return ``True``.

        Parameters
        ----------
        _
            Ignored.

        Returns
        -------
        bool
            Always ``True``.
        """
        return True


def get_transformation(
    class_: str,
    pos: tuple[float, float],
    size: tuple[float, float],
) -> dict[str, str]:
    """Calculate transformation for class.

    The Scaling factor .725, translation constants (6, 5) are arbitrarily
    chosen to fit. Currently only ChoicePseudoState is tranformed.

    Parameters
    ----------
    class_
        Classtype string
    pos
        Position-vector
    size
        Size vector
    """
    if class_ != "ChoicePseudoState":
        return {}

    s = 0.725
    tx, ty = (1 - s) * pos[0] + 6, (1 - s) * pos[1] + 5
    rx, ry = pos[0] + size[0] / 2, pos[1] + size[1] / 2
    return {
        "transform": f"translate({tx},{ty}) scale({s}) rotate(45,{rx},{ry})"
    }



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Common constants and helpers used by all parser submodules."""
from __future__ import annotations

import collections.abc as cabc
import dataclasses
import enum
import logging
import math
import operator
import pathlib
import re
import typing as t

from lxml import builder, etree

import capellambse._namespaces as _n
import capellambse.loader
from capellambse import diagram, helpers

LOGGER: logging.Logger = logging.getLogger(__name__)

ATT_XMID = f"{{{_n.NAMESPACES['xmi']}}}id"
ATT_XMT = f"{{{_n.NAMESPACES['xmi']}}}type"
ATT_XST = f"{{{_n.NAMESPACES['xsi']}}}type"

# Force ports to always have this size.
PORT_SIZE = diagram.Vector2D(10, 10)

RE_COMPOSITE_FILTER = re.compile(r"/@filters\[name='(.*?)'\]$")
RE_STYLECLASS = re.compile(r"/@ownedRepresentations\[name='(.*?)'\]$")
RE_VIEWPOINT = re.compile(r"/@ownedViewpoints\[name='(.*?)'\]$")

XP_ANNOTATION_ENTRIES = etree.XPath(
    (
        "./ownedAnnotationEntries[@source='GMF_DIAGRAMS']"
        "/data[@xmi:type='notation:Diagram']"
    ),
    namespaces=_n.NAMESPACES,
)
XP_DESCRIPTORS = etree.XPath(
    (
        "./ownedRepresentationDescriptors"
        "[@xmi:type='viewpoint:DRepresentationDescriptor']"
    ),
    namespaces=_n.NAMESPACES,
)
XP_VIEWS = etree.XPath(
    "/xmi:XMI/viewpoint:DAnalysis/ownedViews[viewpoint]",
    namespaces=_n.NAMESPACES,
)
ELEMENT = builder.ElementMaker(nsmap={"xmi": str(_n.NAMESPACES["xmi"])})


@dataclasses.dataclass
class ElementBuilder:
    target_diagram: diagram.Diagram
    diagram_tree: etree._Element
    data_element: etree._Element
    melodyloader: capellambse.loader.MelodyLoader
    fragment: pathlib.PurePosixPath


@dataclasses.dataclass
class SemanticElementBuilder(ElementBuilder):
    diag_element: etree._Element
    styleclass: str | None
    melodyobjs: cabc.MutableSequence[etree._Element]


class StackingBox(diagram.Box):
    """A Box with special child-stacking behavior."""

    CHILD_MARGIN = 0

    children: _StackingChildren
    __features: cabc.MutableSequence[str] | None

    def __init__(
        self,
        pos: diagram.Vector2D,
        size: diagram.Vector2D | None = None,
        *,
        stacking_mode: StackingBox._StackingChildren.StackingMode,
        minsize: diagram.Vec2ish = (0, 0),
        maxsize: diagram.Vec2ish = (math.inf, math.inf),
        **kw: t.Any,
    ) -> None:
        del size

        super().__init__(
            pos,
            diagram.Vector2D(0, 0),
            minsize=minsize,
            maxsize=maxsize,
            **kw,
        )
        self.children = self._StackingChildren(self)
        self.stacking_mode = stacking_mode

    def _topsection_size(self) -> diagram.Vector2D:
        """Calculate the size of the top section (this Box' own label).

        Parameters
        ----------
        width
            The maximum width for wrapping label text
        """
        try:
            width = max(map(operator.attrgetter("size.x"), self.children))
        except ValueError:
            width = 0
        pad_x, pad_y = self.padding * 2
        if isinstance(self.label, str):
            label_extent = helpers.get_text_extent(
                self.label, width - pad_x if width else math.inf
            )
        else:
            label_extent = (0, 0)
        width = width or label_extent[0] + pad_x
        if self.features:
            features_height = helpers.get_text_extent(
                "\n".join(self.features), width
            )[1]
            features_height += 24  # Divider line
        else:
            features_height = 0
        height = label_extent[1] + features_height + pad_y
        return diagram.Vector2D(width, height)

    @property
    def size(self) -> diagram.Vector2D:
        pwidth, pheight = self._topsection_size()
        child_bounds = [i.bounds for i in self.children if not i.hidden]
        try:
            width = max(i.size.x for i in child_bounds)
        except ValueError:
            width = pwidth
        cheight = sum(i.size.y for i in child_bounds)
        return diagram.Vector2D(width, cheight + pheight)

    @size.setter
    def size(self, new_size: diagram.Vec2ish) -> None:
        if any(i != 0 for i in new_size):
            raise TypeError("The size of this Box cannot be changed directly")

    @property
    def stacking_mode(self) -> StackingBox._StackingChildren.StackingMode:
        """Return the current stacking mode."""
        return self.children.stacking_mode

    @stacking_mode.setter
    def stacking_mode(self, new_mode: str) -> None:
        try:
            self.children.stacking_mode = self._StackingChildren.StackingMode(
                new_mode
            )
        except ValueError:
            try:
                self.children.stacking_mode = (
                    self._StackingChildren.StackingMode[new_mode]
                )
            except KeyError:
                raise ValueError(
                    f"Invalid stacking mode: {new_mode!r}"
                ) from None

    @property  # type: ignore[override]
    def features(self) -> cabc.Sequence[str] | None:
        """Return the list of Box' features."""
        return self.__features

    @features.setter
    def features(
        self,
        new_features: cabc.Sequence[str] | None,
    ) -> None:
        if new_features is None:
            self.__features = None
        else:
            self.__features = self._SBFeatures(self, new_features)
            try:
                children = self.children
            except AttributeError:
                pass
            else:
                children._restack()

    def snap_to_parent(self, *_: t.Any) -> None:
        pass  # FIXME

    class _SBFeatures(t.MutableSequence[str]):
        def __init__(
            self, parent: StackingBox, *args: t.Any, **kw: t.Any
        ) -> None:
            super().__init__()
            self.__list: list[str] = list(*args, **kw)
            self.__parent = parent

        def __iter__(self) -> cabc.Iterator[str]:
            return iter(self.__list)

        @t.overload
        def __getitem__(self, index: int) -> str: ...

        @t.overload
        def __getitem__(self, index: slice) -> list[str]: ...

        def __getitem__(self, index: int | slice) -> str | list[str]:
            return self.__list[index]

        @t.overload
        def __setitem__(self, index: int, value: str) -> None: ...

        @t.overload
        def __setitem__(
            self, index: slice, value: cabc.Iterable[str]
        ) -> None: ...

        def __setitem__(
            self,
            index: int | slice,
            value: str | cabc.Iterable[str],
        ) -> None:
            self.__list[index] = value  # type: ignore[index,assignment]
            self.__parent.children._restack()

        def __delitem__(self, index: int | slice) -> None:
            del self.__list[index]
            self.__parent.children._restack()

        def __len__(self) -> int:
            return len(self.__list)

        def insert(self, index: int, value: str) -> None:
            self.__list.insert(index, value)
            self.__parent.children._restack()

        def __repr__(self) -> str:
            return repr(self.__list)

    class _StackingChildren(t.MutableSequence[diagram.DiagramElement]):
        class StackingMode(enum.Enum):
            """The possible modes for stacking child boxes."""

            VERTICAL = "VerticalStack"
            HORIZONTAL = "HorizontalStack"

        def __init__(
            self,
            parent: StackingBox,
            stacking_mode: StackingMode = StackingMode.VERTICAL,
        ):
            self.__list: list[diagram.DiagramElement] = []
            self.__parent = parent
            self.stacking_mode = stacking_mode

        def __iter__(self) -> cabc.Iterator[diagram.DiagramElement]:
            return iter(self.__list)

        @t.overload
        def __getitem__(self, index: int) -> diagram.DiagramElement: ...

        @t.overload
        def __getitem__(
            self, index: slice
        ) -> list[diagram.DiagramElement]: ...

        def __getitem__(
            self, index: int | slice
        ) -> diagram.DiagramElement | list[diagram.DiagramElement]:
            return self.__list[index]

        @t.overload
        def __setitem__(
            self, index: int, value: diagram.DiagramElement
        ) -> None: ...

        @t.overload
        def __setitem__(
            self, index: slice, value: cabc.Iterable[diagram.DiagramElement]
        ) -> None: ...

        def __setitem__(
            self,
            index: int | slice,
            value: (
                diagram.DiagramElement | cabc.Iterable[diagram.DiagramElement]
            ),
        ) -> None:
            self.__list[index] = value  # type: ignore[index, assignment]
            self._restack()

        def __delitem__(self, index: int | slice) -> None:
            del self.__list[index]
            self._restack()

        def __len__(self) -> int:
            return len(self.__list)

        def insert(self, index: int, value: diagram.DiagramElement) -> None:
            self.__list.insert(index, value)
            self._restack()

        def _restack(self) -> None:
            if (
                self.stacking_mode
                is StackingBox._StackingChildren.StackingMode.VERTICAL
            ):
                xpos, ypos = self.__parent.pos
                ypos += self.__parent._topsection_size()[1]
                for obj in self.__list:
                    obj_bounds = obj.bounds
                    offset = diagram.Vector2D(xpos, ypos) - obj_bounds.pos
                    obj.move(offset)
                    ypos += obj_bounds.size.y
            elif (
                self.stacking_mode
                is StackingBox._StackingChildren.StackingMode.HORIZONTAL
            ):
                raise NotImplementedError()
            else:
                raise RuntimeError(
                    f"Invalid stacking_mode: {self.stacking_mode!r}"
                )


class CenterAnchoredBox(diagram.Box):
    """A special Box subclass that uses its center as reference point."""

    center: diagram.Vec2Property  # type: ignore[assignment]
    center = diagram.Vec2Property()  # type: ignore[assignment]

    def __init__(
        self, center: diagram.Vec2ish, size: diagram.Vec2ish, **kwargs: t.Any
    ) -> None:
        """Create a CenterAnchoredBox.

        This class supports the same keyword arguments that the parent
        :class:`Box` class supports.

        Parameters
        ----------
        center
            The point at which to place this Box' center
        size
            The width and height of this Box
        """
        super().__init__((math.inf, math.inf), size, **kwargs)
        self.center = center

    @property  # type: ignore[override]
    def pos(self) -> diagram.Vector2D:
        """Return the top left corner position of this Box."""
        return self.center - self.size / 2

    @pos.setter
    def pos(self, new_pos: diagram.Vec2ish) -> None:
        if new_pos == diagram.Vector2D(math.inf, math.inf):
            return
        self.center = diagram.Vector2D(*new_pos) + self.size / 2

    def __repr__(self) -> str:
        return super().__repr__().replace(repr(self.pos), repr(self.center), 1)


class SkipObject(Exception):
    """Raised to stop parsing this object."""

    @classmethod
    def raise_(cls, *args: t.Any, **kw: t.Any) -> t.NoReturn:
        """Raise this exception, ignoring the arguments."""
        del args, kw
        raise cls()


def get_spec_text(seb: SemanticElementBuilder) -> str:
    """Get the ``ownedSpecification`` text for display."""
    try:
        spec = next(seb.melodyobjs[0].iterchildren("ownedSpecification"))
        spec = next(spec.iterchildren("bodies"))
    except StopIteration:
        return ""
    text = helpers.unescape_linked_text(seb.melodyloader, spec.text)
    return text.striptags()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""CLI for the diagram cache updating feature."""
from __future__ import annotations

import pathlib
import sys
import typing as t

import capellambse
from capellambse import _diagram_cache, cli_helpers

# pylint: disable=unused-import
from capellambse._diagram_cache import IndexEntry

try:
    import click
except ImportError:

    def _main() -> None:
        """Display a dependency error."""
        print("Missing dependency 'capellambse[cli]'", file=sys.stderr)
        raise SystemExit(1)

else:

    @click.command()
    @click.option(
        "-m",
        "--model",
        "modelinfo",
        type=cli_helpers.ModelInfoCLI(),
        required=True,
        help="Capella model to export diagrams from",
    )
    @click.option(
        "-o",
        "--output",
        type=click.Path(
            file_okay=False, dir_okay=True, path_type=pathlib.Path
        ),
        default="./diagrams",
        help="Directory to store the rendered diagrams in",
        show_default=True,
    )
    @click.option(
        "-f",
        "--format",
        type=click.Choice(sorted(_diagram_cache.VALID_FORMATS)),
        default="svg",
        help="Image output format",
        show_default=True,
    )
    @click.option(
        "--index/--no-index",
        default=True,
        help="Generate index.json and index.html files",
        show_default=True,
    )
    @click.option(
        "--refresh / --no-refresh",
        default=False,
        help=(
            "Refresh all diagrams before exporting them, to make sure their"
            " contents are up to date with the model. Recommended, but"
            " disabled by default because it may take a long time."
            " NOTE: This may temporarily open a Capella window (see above)."
        ),
    )
    @click.option("--exe", help="Name or path of the Capella executable")
    @click.option("--docker", help="Name of a Docker image containing Capella")
    @click.option(
        "--background/--no-background",
        help="Inserts a white background into the diagrams.",
        default=True,
        show_default=True,
    )
    def _main(
        modelinfo: dict[str, t.Any],
        output: pathlib.Path,
        format: str,
        index: bool,
        exe: str | None,
        docker: str | None,
        background: bool,
        refresh: bool,
    ) -> None:  # noqa: D301
        """Export diagrams from a Capella model.

        This tool can be used to easily populate a diagram cache for use
        with `capellambse.MelodyModel`.

        \b
        Refreshing representations
        --------------------------

        By default, Capella will only automatically refresh diagrams on
        an as-needed basis whenever a diagram is opened in the GUI. This
        can lead to out-of-date images being written by its
        "exportRepresentations" facility, which this tool uses.

        To avoid this problem, the '--refresh' switch can be used to
        automatically refresh all diagrams and make sure they are up to
        date with the model contents.

        Unfortunately, due to the way this refresh operation is
        implemented in Capella, this may cause a Capella window to be
        opened temporarily. It will automatically close once the
        operation is done.

        PLEASE DO NOT CLOSE THE WINDOW that Capella opens during this
        step, or the entire process will be aborted.

        Also be aware that using this option may lead to new objects
        being added in odd places, which may cause severe readability
        issues on the exported diagrams.

        Due to these issues, the automatic refresh is disabled by
        default.
        """
        modelinfo["diagram_cache"] = {"path": output}
        model_ = capellambse.MelodyModel(**modelinfo)

        if docker:
            capella = docker
            force = "docker"
        else:
            capella = exe or "capella{VERSION}"
            force = "exe"

        _diagram_cache.export(
            capella,
            model_,
            format=format,
            index=index,
            force=force,  # type: ignore
            background=background,
            refresh=refresh,
        )


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The abstract FileHandler superclass and utilities."""
from __future__ import annotations

__all__ = [
    "FileHandler",
    "AbstractFilePath",
    "TransactionClosedError",
]

import abc
import collections.abc as cabc
import fnmatch
import importlib.abc as iabc
import os
import pathlib
import typing as t

import typing_extensions as te

from capellambse import helpers

if t.TYPE_CHECKING:
    from capellambse.loader import modelinfo


_F = t.TypeVar("_F", bound="FileHandler")


class FileHandler(metaclass=abc.ABCMeta):
    """Abstract super class for file handler implementations.

    Parameters
    ----------
    path
        The location of the remote. The exact accepted forms are
        determined by the specific file handler implementation, for
        example the ``LocalFileHandler`` accepts only local paths, and
        the ``GitFileHandler`` accepts everything that Git accepts.
    subdir
        Consider all paths relative to this subdirectory, instead of the
        root of the file handler's hierarchy.
    """

    path: str | os.PathLike

    def __init__(
        self,
        path: str | os.PathLike,
        *,
        subdir: str | pathlib.PurePosixPath = "/",
        **kw: t.Any,
    ) -> None:
        super().__init__(**kw)
        self.path = path
        self.subdir = helpers.normalize_pure_path(subdir)

    @abc.abstractmethod
    def get_model_info(self) -> modelinfo.ModelInfo:
        pass

    @abc.abstractmethod
    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        """Open the model file for reading or writing.

        A "file" in this context does not necessarily refer to a
        physical file on disk; it may just as well be streamed in via
        network or other means. Due to this, the file-like returned by
        this method is not required to support random access.

        Parameters
        ----------
        filename
            The name of the file, relative to the ``path`` that was
            given to the constructor.
        mode
            The mode to open the file in. Either ``"r"`` or ``"rb"`` for
            reading, or ``"w"`` or ``"wb"`` for writing a new file. Be
            aware that this method may refuse to open a file for writing
            unless a transaction was started with
            :meth:`write_transaction` first.
        """

    def write_file(
        self, path: str | pathlib.PurePosixPath, content: bytes, /
    ) -> None:
        """Write a file.

        This method is a convenience wrapper around :meth:`open()`.
        """
        with self.open(path, "wb") as f:
            f.write(content)

    def read_file(self, path: str | pathlib.PurePosixPath, /) -> bytes:
        """Read a file.

        This method is a convenience wrapper around :meth:`open()`.
        """
        with self.open(path, "rb") as f:
            return f.read()

    def write_transaction(
        self, **kw: t.Any
    ) -> t.ContextManager[cabc.Mapping[str, t.Any]]:
        """Start a transaction for writing new model files.

        During a transaction, writable objects returned by
        :meth:`open()` buffer their contents in a temporary location,
        and once the transaction ends, all updated files are committed
        to their destinations at once. If the transaction is aborted,
        for example because an exception was raised, then all changes
        must be rolled back to the state immediately before the
        transaction. If, during a transaction, any relevant file is
        touched without the file handler knowing about it, the behavior
        is undefined.

        Note that :meth:`open()` may refuse to open a file as writable
        if no transaction is currently open. This depends on the needs
        of the underlying abstract file system.

        Transaction arguments
        ---------------------
        A concrete file handler implementation may accept arbitrary
        additional arguments to this method. The implementation should
        however always support the case of no arguments given, in which
        case it should start a transaction with sensible defaults, and
        it should also accept and ignore any arguments it does not
        understand. All additional arguments must be passed in via
        keywords. Positional arguments are not supported.

        The return value of the context manager's ``__enter__()`` method
        is expected to be a mapping of all the keyword arguments that
        were not understood. Client code may use this to react properly
        (e.g. by aborting the transaction early) if a required keyword
        argument is found to be not supported by the underlying file
        handler. If a subclass wishes to call its super class'
        ``write_transaction()`` method, it should remove all the keyword
        arguments that it handles itself and pass on the others
        unchanged.

        Well-known arguments
        --------------------
        The following arguments are considered well-known, and their
        meaning is expected to be the same for all file handlers that
        support them.

        -   ``dry_run`` (``bool``): If set to ``True``, changes made
            during the transaction should be rolled back instead of
            being committed, just as if an exception had been raised.
        -   ``author_name`` (``str``): The name of the author of the
            changes.
        -   ``author_email`` (``str``): The e-mail address to record
            alongside the ``author_name``.
        -   ``commit_msg`` (``str``): A message describing the changes,
            which will be recorded in the version control system.
        -   ``remote_branch`` (``str``): If the model came from a remote
            version control system, changes are normally pushed back to
            the same branch on that remote. This argument specifies an
            alternative branch name to push to (which may not yet exist
            on the remote).
        """

        class EmptyTransaction:
            def __enter__(self):
                return kw

            def __exit__(self, *_):
                pass

        return EmptyTransaction()

    @property
    def rootdir(self) -> AbstractFilePath[te.Self]:  # pragma: no cover
        """The root directory of the file handler."""
        raise TypeError(
            f"{type(self).__name__} does not support listing files"
        )

    def iterdir(  # pragma: no cover
        self, path: str | pathlib.PurePosixPath = ".", /
    ) -> cabc.Iterator[AbstractFilePath[te.Self]]:
        """Iterate over the contents of a directory.

        This method is equivalent to calling
        ``fh.rootdir.joinpath(path).iterdir()``.

        Parameters
        ----------
        path
            The directory to list. If not given, lists the contents of
            the root directory (i.e. the one specified by ``path`` and
            ``subdir``).
        """
        raise TypeError(
            f"{type(self).__name__} does not support listing files"
        )


class AbstractFilePath(os.PathLike[str], iabc.Traversable, t.Generic[_F]):
    """A path to a file in a file handler.

    This is an abstract class with FileHandler-agnostic implementations
    of some of Traversable's methods. It is not meant to be instantiated
    directly, but rather to be used as a base class for concrete file
    path implementations.

    Note that some of these implementations may be inefficient, and
    subclasses are encouraged to override them with more efficient
    implementations if possible.
    """

    def __init__(self, parent: _F, path: pathlib.PurePosixPath):
        self._parent = parent
        self._path = path

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._parent!r}, {self._path!r})"

    def __str__(self) -> str:
        return str(self._path)

    def __truediv__(self, path: str | pathlib.PurePosixPath) -> te.Self:
        return self.joinpath(path)

    def __fspath__(self) -> str:
        return str(self._path)

    def joinpath(self, path: str | pathlib.PurePosixPath) -> te.Self:
        newpath = helpers.normalize_pure_path(path, base=self._path)
        return type(self)(self._parent, newpath)

    @abc.abstractmethod
    def is_dir(self) -> bool: ...

    @abc.abstractmethod
    def is_file(self) -> bool: ...

    def iterdir(
        self, path: str | pathlib.PurePosixPath = "."
    ) -> cabc.Iterator[te.Self]:
        path = helpers.normalize_pure_path(path, base=self._path)
        return t.cast("cabc.Iterator[te.Self]", self._parent.iterdir(path))

    @property
    def name(self) -> str:
        return self._path.name

    @property
    def suffix(self) -> str:
        return self._path.suffix

    @property
    def stem(self) -> str:
        return self._path.stem

    @property
    def parent(self) -> te.Self:
        return type(self)(self._parent, self._path.parent)

    def rglob(self, pattern: str) -> cabc.Iterator[te.Self]:
        for f in self.iterdir():
            if fnmatch.fnmatch(f.name, pattern):
                yield f
            if f.is_dir():
                yield from f.rglob(pattern)

    def open(  # type: ignore[override]
        self,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> t.BinaryIO:
        del buffering, errors

        if "b" not in mode:
            raise ValueError("Only binary mode is supported")
        if "r" in mode:
            mode = "rb"
        elif "w" in mode:
            mode = "wb"
        else:
            raise ValueError(f"Unsupported mode: {mode!r}")

        if encoding is not None:
            raise ValueError("Encoding is not supported")
        if newline is not None:
            raise ValueError("Newline is not supported")

        return self._parent.open(self._path, mode)

    def read_bytes(self) -> bytes:
        with self.open("rb") as f:
            return f.read()

    def read_text(self, encoding: str | None = None) -> str:
        with self.open("rb") as f:
            return f.read().decode(encoding or "utf-8")


class TransactionClosedError(RuntimeError):
    """Raised when a transaction must be opened first to write files."""



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import json
import pathlib
import sys

import pytest

import capellambse
from capellambse import aird, diagram, loader

EMPTY_MODEL = pathlib.Path(__file__).parent.joinpath(
    "data/decl/empty_project_52/empty_project_52.aird"
)


class TestAIRDBasicFunctionality:
    test_model = (
        pathlib.Path(__file__).parent / "data" / "parser" / "TestItems.aird"
    )
    test_diagram = "[LAB] Logical System"
    test_json = test_model.with_suffix(".json")
    test_repr = test_model.with_suffix(".repr.txt")
    test_txt = test_model.with_suffix(".txt")

    @pytest.fixture
    def model(self):
        return loader.MelodyLoader(self.test_model)

    @pytest.fixture
    def diagram_under_test(self, model):
        descriptor = next(aird.enumerate_diagrams(model), None)
        assert descriptor is not None
        yield aird.parse_diagram(model, descriptor)

    def test_parsing_all_diagrams_does_not_raise_exceptions(
        self, model, caplog
    ):
        del caplog
        i = 0
        for i, _ in enumerate(aird.parse_diagrams(model), start=1):
            pass
        assert i == 1

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_json_output_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_json.read_text().strip()
        encoder = diagram.DiagramJSONEncoder(indent=4)

        actual = encoder.encode(diagram_under_test)

        assert actual == expected

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_plain_text_representation_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_txt.read_text()
        actual = str(diagram_under_test)
        assert actual + "\n" == expected

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_python_code_representation_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_repr.read_text().strip()
        actual = repr(diagram_under_test)
        assert actual == expected

    def test_enumerate_diagrams_doesnt_crash_if_there_are_no_diagrams(self):
        model = loader.MelodyLoader(EMPTY_MODEL)
        for _ in aird.enumerate_diagrams(model):
            pass


def test_airdparser_msm_produces_valid_json_without_error(
    model: capellambse.MelodyModel,
):
    diag_name = "[MSM] States of Functional Human Being"
    all_diagrams = aird.enumerate_diagrams(model._loader)
    descriptor = next(i for i in all_diagrams if i.name == diag_name)
    parsed = aird.parse_diagram(model._loader, descriptor)

    generated_json = diagram.DiagramJSONEncoder(indent=4).encode(parsed)
    json.loads(generated_json)


@pytest.mark.parametrize(
    ["diag_uid", "num_nodes"],
    [
        ("_7FWu4KrxEeqOgqWuHJrXFA", 34),
        ("_KK2wcKyJEeqCdMaqCWkrKg", 14),
        ("_beibwNYrEeqiU8uzTY0Puw", 11),
        ("_9inmIKgWEeujco-rU7ZOtA", 1),
        ("_Go81cDaWEey5m4gkio8hew", 2),
        ("_-NRNIF2mEey8erljvkS-pQ", 13),
        ("_KZOz4GA9Eey8erljvkS-pQ", 15),
        ("_joc_sGDdEey8erljvkS-pQ", 5),
        ("_bpsckGDhEey8erljvkS-pQ", 4),
        ("_OAg8QHPIEeyW3OIB4qRWZA", 18),
        ("_RSWgcHPIEeyW3OIB4qRWZA", 8),
        ("_bNLx4HPIEeyW3OIB4qRWZA", 6),
        ("_sB8k8Mn-EeyS2Zr7ZWFrXA", 7),
        ("_1adB8GpbEe2Scdb2k-Fvhg", 17),
        ("_fqzi8BKnEeuBCogvtwwNBw", 42),
        ("_ervhAFIqEeyiRNlyKPJwqw", 44),
        ("_8BMtsIQzEeyxv9w6U6_UQg", 6),
        ("_al3zwKg0EeujRPrkuugYGw", 5),
        ("_1sCb8OSREeulYvEGRaHazg", 5),
        ("_APMboAPhEeynfbzU12yy7w", 11),
        ("_K9160IKBEeyd4IW8P6K-AQ", 14),
        ("_0g-esB4cEe2Jgqeci1tAhA", 1),
        ("_hIt8QKWXEeqh-bahrg4jkA", 18),
        ("_u8lIgLB-EeqP7JXhmLvOsg", 26),
        ("_2SjcgAL8EeuogvvqBpYPnQ", 6),
        ("_KTxZ8KJwEeusg-blS2y7tg", 18),
        ("_ZNrYAKJxEeusg-blS2y7tg", 21),
        ("_mKI_kKZwEeuRO5WMG4dh5w", 2),
        ("_MDoM4AWrEeyIAukYgdUX5A", 11),
        ("_VKPRsIQTEeyxv9w6U6_UQg", 26),
        ("_dpUi4LB8EeqLtsTDMgEHBw", 48),
        ("_gRewkLB_EeqLtsTDMgEHBw", 11),
        ("__dDrANZLEeqiU8uzTY0Puw", 14),
        ("_jGNGMOCUEeu-FsE4WGMncQ", 4),
        ("_Ne8zEAATEeykFulkDFvkrg", 7),
        ("_I0nugeQUEe2jpIKbeHI0hw", 14),
    ],
)
def test_iter_visible(model, diag_uid, num_nodes):
    diag = model.diagrams.by_uuid(diag_uid)
    diagram_elements = sum(
        1 for _ in aird.iter_visible(model._loader, diag._element)
    )

    assert diagram_elements == num_nodes



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import collections.abc as cabc
import math
import typing as t

from capellambse import helpers

AlignmentLiteral = t.Literal["center", "left", "right"]


def check_for_horizontal_overflow(
    text: str,
    width: int | float,
    icon_padding: int | float,
    icon_size: int | float,
    alignment: AlignmentLiteral = "center",
) -> tuple[cabc.Sequence[str], float, float]:
    max_text_width = width - (icon_size + 2 * icon_padding)
    assert max_text_width >= 0
    lines = helpers.word_wrap(text, max_text_width)
    text_width = max(w for w, _ in map(helpers.extent_func, lines))
    label_width = text_width + icon_size + 2 * icon_padding
    assert text_width <= max_text_width
    assert label_width <= width
    if alignment == "center":
        label_margin = (width - label_width) / 2
    elif alignment == "left":
        label_margin = 0
    else:
        label_margin = width - label_margin
    return (lines, label_margin, max_text_width)


def check_for_vertical_overflow(
    lines: cabc.Sequence[str],
    height: float | int,
    max_text_width: float | int,
) -> list[str]:
    overflow = ""
    lines_to_render = []
    text_height = 0.0
    for i, (line, (_, line_height)) in enumerate(
        zip(lines, map(helpers.extent_func, lines))
    ):
        if text_height + line_height > height:
            overflow = lines[i - 1] if i else line
            break

        text_height += line_height
        lines_to_render.append(line)

    if overflow:
        line_width = helpers.extent_func(overflow + "...")[0]
        if line_width < max_text_width:
            overflow += "..."
        else:
            dot_width = helpers.extent_func("...")[0]
            new_lines = helpers.word_wrap(
                overflow, math.floor(max_text_width - dot_width)
            )
            overflow = new_lines[0] + "..."

        if lines_to_render:
            lines_to_render[-1] = overflow
        else:
            lines_to_render.append(overflow)

    assert height >= text_height
    return lines_to_render



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The |project| ReqIF extension.

.. deprecated:: 0.5.11

   The ``capellambse.extensions.reqif.elements`` module is deprecated
   and will be removed with capellambse v0.6. Use the classes directly
   from ``capellambse.extensions.reqif`` instead.

   Note that some classes have been renamed to match their equivalent in
   upstream Capella; these are marked as "alias of ..." in the
   documentation below.
"""

from __future__ import annotations

__all__ = [
    "AbstractRequirementsAttribute",
    "AbstractRequirementsRelation",
    "AbstractType",
    "AttributeDefinition",
    "AttributeDefinitionEnumeration",
    "BooleanValueAttribute",
    "DataTypeDefinition",
    "DateValueAttribute",
    "EnumDataTypeDefinition",
    "EnumValue",
    "EnumerationValueAttribute",
    "IntegerValueAttribute",
    "ModuleType",
    "RealValueAttribute",
    "RelationType",
    "RelationsList",
    "ReqIFElement",
    "Requirement",
    "RequirementType",
    "RequirementsFolder",
    "RequirementsIncRelation",
    "RequirementsIntRelation",
    "RequirementsModule",
    "RequirementsOutRelation",
    "RequirementsTypesFolder",
    "StringValueAttribute",
    "XT_FOLDER",
    "XT_INC_RELATION",
    "XT_INT_RELATION",
    "XT_MODULE",
    "XT_MODULE_TYPE",
    "XT_OUT_RELATION",
    "XT_RELATION_TYPE",
    "XT_REQUIREMENT",
    "XT_REQ_ATTRIBUTES",
    "XT_REQ_ATTR_BOOLEANVALUE",
    "XT_REQ_ATTR_DATEVALUE",
    "XT_REQ_ATTR_ENUMVALUE",
    "XT_REQ_ATTR_INTEGERVALUE",
    "XT_REQ_ATTR_REALVALUE",
    "XT_REQ_ATTR_STRINGVALUE",
    "XT_REQ_TYPE",
    "XT_REQ_TYPES",
    "XT_REQ_TYPES_DATA_DEF",
    "XT_REQ_TYPES_F",
    "XT_REQ_TYPE_ATTR_DEF",
    "XT_REQ_TYPE_ATTR_ENUM",
    "XT_REQ_TYPE_ENUM",
    "XT_REQ_TYPE_ENUM_DEF",
    "init",
]

# pylint: disable=unused-import

import collections.abc as cabc
import logging
import os
import re
import typing as t

import markupsafe
from lxml import etree

import capellambse.model
import capellambse.model.common as c
from capellambse.loader import xmltools
from capellambse.model import crosslayer

from . import exporter

XT_REQUIREMENT = "Requirements:Requirement"
XT_REQ_ATTR_STRINGVALUE = "Requirements:StringValueAttribute"
XT_REQ_ATTR_REALVALUE = "Requirements:RealValueAttribute"
XT_REQ_ATTR_INTEGERVALUE = "Requirements:IntegerValueAttribute"
XT_REQ_ATTR_DATEVALUE = "Requirements:DateValueAttribute"
XT_REQ_ATTR_BOOLEANVALUE = "Requirements:BooleanValueAttribute"
XT_REQ_ATTR_ENUMVALUE = "Requirements:EnumerationValueAttribute"
XT_REQ_ATTRIBUTES = {
    XT_REQ_ATTR_ENUMVALUE,
    XT_REQ_ATTR_STRINGVALUE,
    XT_REQ_ATTR_REALVALUE,
    XT_REQ_ATTR_INTEGERVALUE,
    XT_REQ_ATTR_DATEVALUE,
    XT_REQ_ATTR_BOOLEANVALUE,
}
XT_INC_RELATION = "CapellaRequirements:CapellaIncomingRelation"
XT_OUT_RELATION = "CapellaRequirements:CapellaOutgoingRelation"
XT_INT_RELATION = "Requirements:InternalRelation"
XT_MODULE = "CapellaRequirements:CapellaModule"
XT_FOLDER = "Requirements:Folder"

XT_REQ_TYPES_F = "CapellaRequirements:CapellaTypesFolder"
XT_REQ_TYPES_DATA_DEF = "Requirements:DataTypeDefinition"
XT_REQ_TYPE = "Requirements:RequirementType"
XT_RELATION_TYPE = "Requirements:RelationType"
XT_MODULE_TYPE = "Requirements:ModuleType"
XT_REQ_TYPE_ENUM = "Requirements:EnumerationDataTypeDefinition"
XT_REQ_TYPE_ATTR_ENUM = "Requirements:EnumValue"
XT_REQ_TYPE_ATTR_DEF = "Requirements:AttributeDefinition"
XT_REQ_TYPE_ENUM_DEF = "Requirements:AttributeDefinitionEnumeration"
XT_REQ_TYPES = {
    XT_REQ_TYPES_F,
    XT_REQ_TYPES_DATA_DEF,
    XT_REQ_TYPE,
    XT_RELATION_TYPE,
    XT_MODULE_TYPE,
    XT_REQ_TYPE_ENUM,
    XT_REQ_TYPE_ATTR_ENUM,
    XT_REQ_TYPE_ATTR_DEF,
    XT_REQ_TYPE_ENUM_DEF,
}

logger = logging.getLogger("reqif")

from . import (
    AbstractRequirementsAttribute,
    AbstractRequirementsRelation,
    AbstractType,
    AttributeDefinition,
    AttributeDefinitionEnumeration,
    BooleanValueAttribute,
)
from . import CapellaIncomingRelation as RequirementsIncRelation
from . import CapellaModule as RequirementsModule
from . import CapellaOutgoingRelation as RequirementsOutRelation
from . import CapellaTypesFolder as RequirementsTypesFolder
from . import DataTypeDefinition, DateValueAttribute
from . import EnumerationDataTypeDefinition as EnumDataTypeDefinition
from . import EnumerationValueAttribute, EnumValue
from . import Folder as RequirementsFolder
from . import IntegerValueAttribute
from . import InternalRelation as RequirementsIntRelation
from . import (
    ModuleType,
    RealValueAttribute,
    RelationsList,
    RelationType,
    ReqIFElement,
    Requirement,
    RequirementType,
    StringValueAttribute,
    init,
)
